import pathlib
from fnmatch import fnmatch
from functools import singledispatch
from typing import Optional

import pandas as pd

from footings.utils import dispatch_function

from .io import load_footings_json_file, load_footings_xlsx_file
from .io.to_xlsx import FootingsXlsxEntry

#########################################################################################
# comparison functions
#########################################################################################

MAX_STR_LEN_TO_SHOW = 100


@singledispatch
def compare_values(result_value, expected_value, tolerance=None):
    """A dispatch function to compare two values."""
    return compare_values(str(result_value), str(expected_value))


@compare_values.register(str)
def _(result_value, expected_value, tolerance=None):
    if type(result_value) != type(expected_value):
        return compare_values(result_value, str(expected_value), tolerance)
    result = result_value == expected_value
    if result:
        msg = ""
    else:
        if (
            len(result_value) <= MAX_STR_LEN_TO_SHOW
            and len(expected_value) <= MAX_STR_LEN_TO_SHOW
        ):
            msg = f"The result value [{result_value}] is not equal to the expected value [{expected_value}]."
        else:
            msg = "The result and expected strings are not equal and they are to long to display."
    return result, msg


@compare_values.register(bool)
@compare_values.register(int)
@compare_values.register(float)
def _(result_value, expected_value, tolerance=None):
    if type(result_value) != type(expected_value):
        return compare_values(str(result_value), str(expected_value))
    if result_value == float("nan") or expected_value == float("nan"):
        return compare_values(str(result_value), str(expected_value))
    if tolerance is None:
        result = result_value == expected_value
    else:
        result = abs(result_value - expected_value) < tolerance
    if result:
        msg = ""
    else:
        msg = f"The result value [{str(result_value)}] is not equal to the expected value [{str(expected_value)}]."
        if tolerance is not None:
            msg = msg[:-1] + f" using a tolerance of {str(tolerance)}."
    return result, msg


@compare_values.register(list)
def _(result_value, expected_value, tolerance=None):
    result = all(
        compare_values(r, e, tolerance)[0] for r, e in zip(result_value, expected_value)
    )
    if result:
        msg = ""
    else:
        msg = "The result list values are different from the expected list values.\n\n"
        msg += f"The result list values are - \n\n{str(result_value)}\n\n"
        msg += f"The expected list values are - \n\n{str(expected_value)}\n\n"
    return result, msg


@compare_values.register(pd.Series)
def _(result_value, expected_value, tolerance=None):
    result = all(
        compare_values(r, e, tolerance)[0] for r, e in zip(result_value, expected_value)
    )
    if result:
        msg = ""
    else:
        if result_value.name is None:
            msg = "The result pd.Series is different from the expected pd.Series.\n\n"
        else:
            msg = f"The result pd.Series for column [{result_value.name}] is different"
            msg += " from the expected pd.Series.\n\n"
        msg += f"The result pd.Series is - \n\n{str(result_value.values)}\n\n"
        msg += f"The expected pd.Series is - \n\n{str(expected_value.values)}\n\n"
    return result, msg


@compare_values.register(pd.DataFrame)
def _(result_value, expected_value, tolerance=None):
    series_compare = [
        compare_values(result_value[r], expected_value[e], tolerance)
        for r, e in zip(result_value.columns, expected_value.columns)
    ]
    result = all(x[0] for x in series_compare)
    if result:
        msg = ""
    else:
        msg = "The result pd.DataFrame is different from the expected pd.DataFrame.\n\n"
        msg += "\n\n".join(x[1] for x in series_compare)
    return result, msg


def make_key_checker(exclude_keys: Optional[list] = None) -> callable:
    if exclude_keys is None:
        return lambda key: False

    def key_checker(key):
        if isinstance(key, FootingsXlsxEntry):
            key = key.mapping
        return any(fnmatch(key, x) for x in exclude_keys)

    return key_checker


def compare_file_dicts(
    result: dict,
    expected: dict,
    tolerance: Optional[float],
    exclude_keys: Optional[list],
):
    """ """
    test = True
    log = {}
    key_exclude = make_key_checker(exclude_keys)
    keys = set(list(result.keys()) + list(expected.keys()))
    for key in keys:
        temp = True
        if key_exclude(key):
            continue
        try:
            res_val = result[key]
        except KeyError:
            msg = (
                "The result file is missing this key, but it exists in the expected file."
            )
            temp = False

        try:
            exp_val = expected[key]
        except KeyError:
            msg = (
                "The expected file is missing this key, but it exists in the result file."
            )
            temp = False

        if temp:
            if type(res_val) != type(exp_val):
                temp = False
                msg = "The value types are different."
            else:
                temp, msg = compare_values(res_val, exp_val, tolerance)

        if temp is False:
            test = False
        log.update({key: {"result": temp, "msg": msg}})

    message = "\n".join(
        [f"{k} : {v['msg']}" for k, v in log.items() if v["result"] is False]
    )
    return test, message


def check_extensions_equal(result, expected):
    result_ext = pathlib.Path(result).suffix
    expected_ext = pathlib.Path(expected).suffix
    if result_ext != expected_ext:
        msg = f"The file extensions for result [{result_ext}] and expected [{expected_ext}] do not match."
        raise ValueError(msg)
    return True


def assert_footings_json_files_equal(
    result: str,
    expected: str,
    tolerance: Optional[float] = None,
    exclude_keys: Optional[list] = None,
):
    """Assert whether two footing json files are equal.  This function is
    useful for unit testing models to ensure models stay true over time.

    Parameters
    ----------
    result : str
        The path to the result file.
    expected : str
        The path to the expected file.
    tolerance : float, optional
        The tolerance to test on numeric values.
    exclude_keys : list, optional
        Keys to exclude from testing.

    Returns
    -------
    bool
        True or false on whether files are equal given parameters.

    Raises
    ------
    ValueError
        If the result and expected files share different extension types.
    AssertionError
        If any records between the result and expected files are different.
    """
    check_extensions_equal(result, expected)
    result = load_footings_json_file(result)
    expected = load_footings_json_file(expected)
    test, message = compare_file_dicts(
        result=result, expected=expected, tolerance=tolerance, exclude_keys=exclude_keys,
    )
    if test is False:
        raise AssertionError(f"\n{str(message)}")
    return True


def assert_footings_xlsx_files_equal(
    result: str,
    expected: str,
    tolerance: Optional[float] = None,
    exclude_keys: Optional[list] = None,
):
    """Assert whether two footing xlsx files are equal. This function is
    useful for unit testing models to ensure models stay true over time.

    Parameters
    ----------
    result : str
        The path to the result file.
    expected : str
        The path to the expected file.
    tolerance : float, optional
        The tolerance to test on numeric values.
    exclude_keys : list, optional
        Keys to exclude from testing.

    Returns
    -------
    bool
        True or false on whether files are equal given parameters.

    Raises
    ------
    ValueError
        If the result and expected files share different extension types.
    AssertionError
        If any records between the result and expected files are different.
    """
    check_extensions_equal(result, expected)
    result = load_footings_xlsx_file(result)
    expected = load_footings_xlsx_file(expected)
    test, message = compare_file_dicts(
        result=result, expected=expected, tolerance=tolerance, exclude_keys=exclude_keys,
    )
    if test is False:
        raise AssertionError(f"\n{message}")
    return True


@dispatch_function(key_parameters=("file_ext",))
def _assert_footings_files_equal(file_ext, result, expected, **kwargs):
    """test run_model audit"""
    msg = "No registered function based on passed paramters and no default function."
    raise NotImplementedError(msg)


@_assert_footings_files_equal.register(file_ext=".json")
def _(result: str, expected: str, **kwargs):
    assert_footings_json_files_equal(result, expected, **kwargs)


@_assert_footings_files_equal.register(file_ext=".xlsx")
def _(result: str, expected: str, **kwargs):
    assert_footings_xlsx_files_equal(result, expected, **kwargs)


def assert_footings_files_equal(
    result: str,
    expected: str,
    tolerance: Optional[float] = None,
    exclude_keys: Optional[list] = None,
):
    """A generic function to assert whether two footing files are equal. This function is
    useful for unit testing models to ensure models stay true over time.

    Parameters
    ----------
    result : str
        The path to the result file.
    expected : str
        The path to the expected file.
    tolerance : float, optional
        The tolerance to test on numeric values.
    exclude_keys : list, optional
        Keys to exclude from testing.

    Returns
    -------
    bool
        True or false on whether files are equal given parameters.

    Raises
    ------
    ValueError
        If the result and expected files share different extension types.
    AssertionError
        If any records between the result and expected files are different.

    See Also
    --------
    assert_footings_json_files_equal
    assert_footings_xlsx_files_equal
    """
    file_ext = pathlib.Path(result).suffix
    _assert_footings_files_equal(
        file_ext=file_ext,
        result=result,
        expected=expected,
        tolerance=tolerance,
        exclude_keys=exclude_keys,
    )



import inspect
import types
from collections.abc import Iterable
from functools import partial, update_wrapper, wraps
from itertools import product
from typing import Callable, Tuple

__all__ = ["dispatch_function", "once"]


class DispatchMissingArgumentError(ValueError):
    """Missing Argument to dispatch_function.register"""


def dispatch_function(key_parameters: Tuple, default_function: Callable = None):
    """Transform a function into a generic dispatch function.

    A dispatch function operates similar to `functools.singledispatch` where a generic function
    is recorded and additional functions can be registered and called based on passed arguments.
    However, `functools.singledispatch` calls a registered function based on the type of the first
    argument. Dispatch function calls a registered function based on the key word arguments assigned
    when registering a function. If a registered function does not exist for the passed parameters,
    the default functions is called. To force only registerd functions to be used, have the
    default function `raise NotImplementedError`.

    A dispatch_function is a useful function to replace a block of if else statements that call
    different functions.

    Parameters
    ----------
    key_parameters : tuple
        A tuple of the paramters to use as
    default_function : callable
        The default function to use when the function is called and a regestered function cannot be
        found. function docstring and signature sourced.

    See Also
    --------
    functools.singledispatch

    Examples
    --------
    >>> @dispatch_function(key_parameters=("key",))
    >>> def dispatch(key):
    >>>     return "default"
    >>>
    >>> @dispatch.register(key="x")
    >>> def _():
    >>>     return "x"
    >>>
    >>> @dispatch.register(key="y")
    >>> def _():
    >>>     return "y"
    >>>
    >>> assert dispatch(key="x") == "x"
    >>> assert dispatch(key="y") == "y"
    >>> assert dispatch(key="z") == "default"
    >>> assert dispatch(key=None) == "default"
    """
    if default_function is None:
        return partial(dispatch_function, key_parameters)

    registry = {}

    def register(function: Callable = None, **kwargs):
        """Register a function using a key established by the required parameters.

        Parameters
        ----------
        function : callable
            The function to register.
        **kwargs
            The key_parameters to register for the given function.

        Raises
        ------
        DispatchMissingArgumentError
            Missing Argument to dispatch_function.register.
        """
        if function is None:
            return partial(register, **kwargs)

        items = []
        for key in key_parameters:
            value = kwargs.get(key, None)
            if value is None:
                msg = f"The parameter [{key}] is not a parameter of the instance."
                raise DispatchMissingArgumentError(msg)
            if isinstance(value, str):
                items.append([value])
            elif isinstance(value, Iterable):
                items.append(value)
            else:
                items.append([value])
        keys = list(product(*items))
        for key in keys:
            registry.update({key: function})

    def wrapper(**kwargs):
        orig_kwargs = kwargs.copy()
        key_args = []
        for key in key_parameters:
            try:
                key_args.append(kwargs.get(key))
                kwargs.pop(key)
            except KeyError:
                pass
        key_args = tuple(key_args)
        registry_function = registry.get(key_args, None)
        if registry_function is None:
            return default_function(**orig_kwargs)
        return registry_function(**kwargs)

    wrapper.register = register
    wrapper.registry = types.MappingProxyType(registry)
    wrapper.__signature__ = inspect.signature(default_function)
    update_wrapper(wrapper, default_function)
    return wrapper


def loaded_function(function: Callable = None):
    """Transform a function into a generic loaded function.

    A loaded function is an object that allows hooks to be registered and called either before
    or after the function is called. Hooks are registered as functions and need a positional
    argument of 'start' to call before the function or 'end' to be call after the function.

    This is a useful object to add validation of inputs before the function is called or validation
    of output after the function has been called.

    Parameters
    ----------
    function: callable
        The primary function.

    Examples
    --------
    >>> @loaded_function
    >>> def loaded(a, b):
    >>>     '''Main function'''
    >>>     return a + b
    >>>
    >>> @loaded.register(position="start")
    >>> def _(a, b):
    >>>     b += 1
    >>>     return a, b
    >>>
    >>> @loaded.register(position="end")
    >>> def _(x):
    >>>     return x + 1
    >>>
    >>> with pytest.raises(ValueError):
    >>>
    >>>     @loaded.register(position="zzz")
    >>>     def _(x):
    >>>         pass
    >>>
    >>> assert loaded(a=1, b=1) == 4
    >>> assert loaded(1, 1) == 4
    """
    if function is None:
        return partial(loaded_function)

    # test if function is already loaded
    if hasattr(function, "loaded"):
        loaded = function.loaded
        need_to_register = False
    else:
        loaded = []
        need_to_register = True

    def register(position: str, function: Callable = None):
        """Register a function into the registry.

        Parameters
        ----------
        function : callable
            The function to register.
        position : str
            Use 'end' if adding to the end of the registry or 'start' to add to the begining.

        Raises
        ------
        ValueError
            If position is not 'end' or 'start'.
        """
        if function is None:
            return partial(register, position)
        if position == "end":
            loaded.append(function)
        elif position == "start":
            loaded.insert(0, function)
        else:
            msg = f"The position {position} is not known. Please use 'start' or 'end'."
            raise ValueError(msg)

    def wrapper(*args, **kwargs):
        ret = loaded[0](*args, **kwargs)
        if len(loaded) > 1:
            for func in loaded[1:]:
                if isinstance(ret, tuple):
                    ret = func(*ret)
                else:
                    ret = func(ret)
        return ret

    wrapper.register = register
    wrapper.loaded = loaded
    wrapper.__signature__ = inspect.signature(function)
    update_wrapper(wrapper, function)
    if need_to_register is True:
        wrapper.register(position="end", function=function)
    return wrapper


# https://discuss.python.org/t/reduce-the-overhead-of-functools-lru-cache-for-functions-with-no-parameters/3956/3
def once(function):
    obj = None

    @wraps(function)
    def inner():
        nonlocal obj
        if obj is None:
            obj = function()
        return obj

    return inner



import functools
import operator
from typing import Optional, Union

import numpy as np
import pandas as pd


def calc_continuance(
    *decrements: pd.Series, starting_duration: Optional[pd.Series] = None,
) -> pd.Series:
    """Calculate continuance given a set of decrements.

    Parameters
    ----------
    *decrements : pd.Series
        The series of decrements to apply to calculate continuance.
    starting_duration : pd.Series, optional
        The start duration continuance value.

    Returns
    -------
    pd.Series
        The continuance table.

    Examples
    --------
    >>> import pandas as pd
    >>> from footings.actuarial_tools import calc_continuance
    >>> mortality_rate = pd.Series([0.01, 0.015, 0.02])
    >>> lapse_rate = pd.Series([0.2, 0.1, 0.05])

    Assume -
    - mortality is applied uniformly over a duration
    - lapse occurs only at end of duration

    Calculate continuance at the end of a duration.
    >>> lives_ed = calc_continuance(mortality_rate, lapse_rate)
    >>> lives_ed
    0   0.792000
    1   0.702108
    2   0.653663

    Shift one place to get the beginning duration value.
    >>> lives_bd = lives_ed.shift(1, fill_value=1)
    >>> lives_bd
    0   1.000000
    1   0.792000
    2   0.702108

    To get the value for the mid-point of the duration set a starting_duration value.
    >>> lives_md = calc_continuance(mortality_rate / 2, starting_duration=lives_bd)
    >>> lives_md
    0   0.995000
    1   0.786060
    2   0.695087
    """
    if starting_duration is None:
        return functools.reduce(operator.mul, [1 - d for d in decrements]).cumprod()
    return starting_duration * functools.reduce(operator.mul, [1 - d for d in decrements])


def calc_discount(
    interest_rate: pd.Series, *, t_adj: Optional[Union[int, float]] = None
) -> pd.Series:
    """Calculate the discount factor over a series.

    Parameters
    ----------
    interest_rate : pd.Series
        A series of interest rate values
    t_adj : Union[int, float], optional
        An optinal adjustment to the timing for the current duration. As an example, use 0.5 for a midpoint.

    Returns
    -------
    pd.Series
        The discount factor.

    Examples
    --------
    >>> import pandas as pd
    >>> from footings.actuarial_tools import calc_discount
    >>> interest_rate = pd.Series([0.03, 0.04, 0.05])
    >>> v_ed = calc_discount(interest_rate)
    >>> v_ed
    0    0.970874
    1    0.933532
    2    0.889079

    >>> v_md = calc_discount(interest_rate, t_adj=0.5)
    >>> v_md
    0    0.985329
    1    0.952020
    2    0.911034

    >>> v_bd = calc_discount(interest_rate, t_adj=0)
    >>> v_bd
    0    1.000000
    1    0.970874
    2    0.933532
    """
    cum = (1 + interest_rate).cumprod()
    if t_adj is not None:
        cum = cum.shift(1, fill_value=1) * (1 + interest_rate) ** t_adj
    return 1 / cum


def calc_interpolation(
    val_0: pd.Series,
    val_1: pd.Series,
    wt_0: pd.Series,
    wt_1: Optional[pd.Series] = None,
    method: str = "linear",
) -> pd.Series:
    """Calculate interpolation between two values.

    Parameters
    ----------
    val_0 : pd.Series
        zzz
    val_1 : pd.Sereis
        zzz
    wt_0 : pd.Series
        zzz
    wt_1 : pd.Series, optional
        zzz
    method : str
        zzz

    Returns
    -------
    pd.Series
        The interpolated values.

    Examples
    --------
    >>> import pandas as pd
    >>> from footings.actuarial_tools import calc_interpolation
    >>> val_0 = pd.Series([1, 2, 3])
    >>> val_1 = pd.Series([2, 3, 4])
    >>> wt_0 = pd.Series([0.5, 0.5, 0.5])
    >>> linear = calc_interpolation(val_0, val_1, wt_0, method="linear")
    >>> linear
    0    1.5
    1    2.5
    2    3.5
    >>> log = calc_interpolation(val_0, val_1, wt_0, method="log")
    >>> log
    0    1.414214
    1    2.449490
    2    3.464102
    """
    if wt_1 is None:
        wt_1 = 1 - wt_0

    if method == "linear":
        ret = val_0 * wt_0 + val_1 * wt_1
    elif method == "log":
        ret = np.exp(np.log(val_1) * wt_1 + np.log(val_0) * wt_0)
    else:
        msg = f"The value passed to method [{str(method)}] is not recognized."
        msg += " Please use one of 'linear' or 'log'."
        raise ValueError(msg)
    return ret


def calc_pv(series: pd.Series) -> pd.Series:
    """Calculate the present value of a series.

    Parameters
    ----------
    series : pd.Series
        The series to calculate the present value from.

    Returns
    -------
    pd.Series
        The series with the present value for each row.

    Examples
    --------
    >>> import pandas as pd
    >>> from footings.actuarial_tools import calc_pv
    >>> x = calc_pv(pd.Series([3, 2, 1]))
    0    6
    1    3
    2    1
    """
    return series[::-1].cumsum()[::-1]


def calc_pvfnb(pvfb: pd.Series, pvfp: pd.Series, net_benefit_method: str) -> pd.Series:
    """Calculate the present value of net future benefits.

    Parameters
    ----------
    pvfb : pd.Series
        A series representing the present value of future benefits.
    pvfp : pd.Series
        A series representing the present value of premium future premium.
    net_benefit_method : str
        The net benefit method to use. Set to  -
        - NLP for net level premium
        - PT1 for 1 year preliminary term
        - PT2 for 2 year preliminary term

    Returns
    -------
    pd.Series
        A series representing the present value of net future benefits.

    Examples
    --------
    >>> import pandas as pd
    >>> from footings.actuarial_tools import calc_pvfnb
    >>> pvfb = pd.Series([6, 5, 3])
    >>> pvfp = pd.Series([9, 6, 3])
    >>> pvfnb = calc_pvfnb(pvfb=pvfb, pvfp=pvfp, net_benefit_method="NLP")
    >>> pvfnb
    0    6.0
    1    4.0
    2    2.0
    """
    if net_benefit_method == "NLP":
        nlp = pvfb.iat[0] / pvfp.iat[0]
        pvfnb = pvfp * nlp
    elif net_benefit_method == "PT1":
        nlp = pvfb.iat[1] / pvfp.iat[1]
        pvfnb = pvfp * nlp
        pvfnb.iat[0] = pvfb.iat[0]
    elif net_benefit_method == "PT2":
        nlp = pvfb.iat[2] / pvfp.iat[2]
        pvfnb = pvfp * nlp
        pvfnb.iat[0] = pvfb.iat[0]
        pvfnb.iat[1] = pvfb.iat[1]
    else:
        msg = f"The net_benefit_method of [{net_benefit_method}] is not recognized."
        msg += " Use one of NLP, PT1, or PT2."
        raise ValueError(msg)
    return pvfnb


def calc_benefit_reserve(
    pvfb: pd.Series, pvfnb: pd.Series, lives: pd.Series, discount: pd.Series,
) -> pd.Series:
    """Calculate benefit reserve.

    Parameters
    ----------
    pvfb : pd.Series
        A series representing the present value of future benefits.
    pvfnb : pd.Series
        A series representing the present value of future net benefits.
    lives : pd.Series
        A series representing the expected lives inforce.
    discount : pd.Series
        A series representing the discount factor.

    Returns
    -------
    pd.Series
        The benefit reserve over time.

    Examples
    --------
    >>> import pandas as pd
    >>> from footings.actuarial_tools import calc_pvfnb, calc_benefit_reserve
    >>> pvfb = pd.Series([6, 5, 3])
    >>> pvfp = pd.Series([9, 6, 3])
    >>> pvfnb = calc_pvfnb(pvfb=pvfb, pvfp=pvfp, net_benefit_method="NLP")
    >>> lives = pd.Series([0.95, 0.9, 0.8])
    >>> discount = pd.Series([0.95, 0.9, 0.85])
    >>> calc_benefit_reserve(
    ...     pvfb=pvfb,
    ...     pvfnb=pvfnb,
    ...     lives=lives,
    ...     discount=discount,
    ... )
    0    1.108033
    1    1.234568
    2    0.000000
    """
    return (pvfb - pvfnb).shift(-1, fill_value=0) / lives / discount


# def calc_change_in_reserve(
#     reserve: pd.Series, lives: pd.Series, discount: pd.Series
# ):
#     """Calculate change in reserve."""
#     discount_shift = discount.shift(1, fill_value=1)
#     current_period = reserve * lives / (discount_shift / discount)
#     prior_period = reserve.shift(1, fill_value=0) * lives.shift(1, fill_value=1)
#     return current_period - prior_period


# def calc_policy_year_benefit_reserve(
#     gross_premiums: pd.Series,
#     benefits: pd.Series,
#     lives: pd.Series,
#     discount: pd.Series,
#     net_benefit_method: str,
# ):
#     """Calculate a policy year benefit reserve.
#
#     Parameters
#     ----------
#     gross_premiums : pd.Series
#         The projected stream  of gross premiums.
#     benefits : pd.Series
#         The projected stream of benefits.
#     lives : pd.Series
#         The projected lives inforce at each duration.
#     discount : pd.Series
#         The discount to apply.
#     net_benefit_method : str
#         The net benefit method to use. Options are -
#             - NLP (Net Level Premium)
#             - PT1 (1 Year Preliminary Term)
#             - PT2 (2 Year Preliminary Term)
#
#     Returns
#     -------
#     pd.Series
#         A series with the benefit reserves.
#     """
#     pvfb = calc_pv(benefits * lives * discount)
#     pvp = calc_pv(gross_premiums * lives * discount)
#     pvfnb = calc_pvfnb(pvfb, pvp, net_benefit_method)
#     benefit_reserve = (pvfb - pvfnb) / discount
#     return benefit_reserve



"""A Sphinx extension that uses and builds upon the Sphinx extension numpydoc_ to build custom documentation for models built using the Footings framework.

Documentation for a Footings based model includes sections for Parameters, Sensitivitys, Meta, Intermediates, Returns, and Steps as well as the standard sections of the numpy docstring format.

To use this extension, include ``footings.doctools`` as an extension with your conf.py file.

.. code-block:: python

   extensions = [
       "footings.doctools",
       ...
    ]

.. _numpydoc: https://numpydoc.readthedocs.io/en/latest/
"""


def setup(app, *args, **kwargs):
    from numpydoc.numpydoc import setup

    from .docscrape_sphinx import get_doc_object

    return setup(app, *args, get_doc_object_=get_doc_object, **kwargs)



import inspect
import os
import pydoc
from collections.abc import Callable

from jinja2 import FileSystemLoader
from jinja2.sandbox import SandboxedEnvironment
from numpydoc.docscrape_sphinx import (
    SphinxClassDoc,
    SphinxDocString,
    SphinxFunctionDoc,
    SphinxObjDoc,
)
from sphinx.jinja2glue import BuiltinTemplateLoader

from .docscrape import FootingsDoc


class SphinxFootingsDoc(SphinxDocString, FootingsDoc):
    def __init__(self, obj, doc=None, func_doc=None, config={}):
        self.load_config(config)
        FootingsDoc.__init__(self, obj, doc=doc, func_doc=None, config=config)

    def load_config(self, config):
        self.use_plots = config.get("use_plots", False)
        self.use_blockquotes = config.get("use_blockquotes", False)
        self.class_members_toctree = config.get("class_members_toctree", True)
        self.attributes_as_param_list = config.get("attributes_as_param_list", True)
        self.xref_param_type = config.get("xref_param_type", False)
        self.xref_aliases = config.get("xref_aliases", dict())
        self.xref_ignore = config.get("xref_ignore", set())
        self.template = config.get("template", None)
        if self.template is None:
            template_dirs = [os.path.join(os.path.dirname(__file__), "templates")]
            template_loader = FileSystemLoader(template_dirs)
            template_env = SandboxedEnvironment(loader=template_loader)
            self.template = template_env.get_template("footings_docstring.rst")

    def _str_steps(self):
        """Generate RST for steps"""
        out = []
        if self["Steps"]:
            out += self._str_field_list("Steps")
            out += [""]
            for step in self["Steps"]:
                out += self._str_indent([step])
            out += [""]
        return out

    def __str__(self, indent=0, func_role="obj"):
        ns = {
            "signature": self._str_signature(),
            "index": self._str_index(),
            "summary": self._str_summary(),
            "extended_summary": self._str_extended_summary(),
            "parameters": self._str_param_list("Parameters"),
            "sensitivities": self._str_param_list("Sensitivities"),
            "meta": self._str_param_list("Meta"),
            "intermediates": self._str_param_list("Intermediates"),
            "returns": self._str_param_list("Returns"),
            "yields": self._str_returns("Yields"),
            "receives": self._str_returns("Receives"),
            "other_parameters": self._str_param_list("Other Parameters"),
            "steps": self._str_steps(),
            "raises": self._str_returns("Raises"),
            "warns": self._str_returns("Warns"),
            "warnings": self._str_warnings(),
            "see_also": self._str_see_also(func_role),
            "notes": self._str_section("Notes"),
            "references": self._str_references(),
            "examples": self._str_examples(),
            "methods": self._str_member_list("Methods"),
        }

        ns = dict((k, "\n".join(v)) for k, v in ns.items())
        rendered = self.template.render(**ns)
        return "\n".join(self._str_indent(rendered.split("\n"), indent))


def get_doc_object(obj, what=None, doc=None, config={}, builder=None):

    if what is None:
        if inspect.isclass(obj):
            what = "class"
        elif inspect.ismodule(obj):
            what = "module"
        elif isinstance(obj, Callable):
            what = "function"
        else:
            what = "object"

    if what == "class" and hasattr(obj, "run") and hasattr(obj, "audit"):
        what = "footing"
    template_dirs = [os.path.join(os.path.dirname(__file__), "templates")]
    if builder is not None:
        template_loader = BuiltinTemplateLoader()
        template_loader.init(builder, dirs=template_dirs)
    else:
        template_loader = FileSystemLoader(template_dirs)
    template_env = SandboxedEnvironment(loader=template_loader)
    config["template"] = template_env.get_template("footings_docstring.rst")
    if what == "footing":
        return SphinxFootingsDoc(obj, func_doc=SphinxFunctionDoc, doc=doc, config=config)
    elif what == "class":
        return SphinxClassDoc(obj, func_doc=SphinxFunctionDoc, doc=doc, config=config)
    elif what in ("function", "method"):
        return SphinxFunctionDoc(obj, doc=doc, config=config)
    else:
        if doc is None:
            doc = pydoc.getdoc(obj)
        return SphinxObjDoc(obj, doc, config=config)



import pathlib
from copy import deepcopy
from inspect import getfullargspec, signature

from attr import asdict, attrib, attrs
from attr.validators import instance_of, optional

from .io import create_footings_json_file, create_footings_xlsx_file
from .utils import dispatch_function


def _make_signature(model):
    sig = model.__class__.__qualname__
    sig += str(signature(model.__class__))
    return sig[: (sig.find(")") + 1)]


@attrs(slots=True, frozen=True)
class AuditStepConfig:
    show_method_name = attrib(type=bool, default=True)
    show_docstring = attrib(type=bool, default=True)
    show_uses = attrib(type=bool, default=True)
    show_impacts = attrib(type=bool, default=True)
    show_output = attrib(type=bool, default=True)
    show_metadata = attrib(type=bool, default=True)


@attrs(slots=True, frozen=True)
class AuditConfig:
    show_signature = attrib(type=bool, default=True)
    show_docstring = attrib(type=bool, default=True)
    show_steps = attrib(type=bool, default=True)
    step_config = attrib(type=AuditStepConfig, default=AuditStepConfig())


def _get_model_output(model):
    steps = []
    for step in model.__model_steps__:
        method = getattr(model, step)
        step_obj = method.func
        method()
        results = deepcopy(model)
        if hasattr(model, step + "_audit"):
            getattr(results, step + "_audit")()
        step_output = {
            item: getattr(results, item.split(".")[1]) for item in step_obj.impacts
        }
        steps.append(step_output)

    final_output = {ret: getattr(model, ret) for ret in model.__model_returns__}

    return steps, final_output


@attrs(slots=True, frozen=True)
class AuditStepContainer:
    name = attrib(type=str)
    method_name = attrib(type=str, default=None)
    docstring = attrib(type=str, default=None)
    uses = attrib(type=tuple, default=None)
    impacts = attrib(type=tuple, default=None)
    output = attrib(type=dict, default=None)
    metadata = attrib(type=dict, default=None)
    config = attrib(
        type=AuditStepConfig,
        factory=AuditStepConfig,
        validator=instance_of(AuditStepConfig),
    )

    @classmethod
    def create(cls, step, output, config):
        kwargs = {"name": step.name, "config": config}
        if config.show_method_name:
            kwargs.update({"method_name": step.method_name})
        if config.show_docstring:
            kwargs.update({"docstring": step.docstring})
        if config.show_uses:
            kwargs.update({"uses": step.uses})
        if config.show_impacts:
            kwargs.update({"impacts": step.impacts})
        if config.show_output:
            kwargs.update({"output": output})
        if config.show_metadata:
            kwargs.update({"metadata": step.metadata})
        return cls(**kwargs)

    def as_dict(self):
        d = asdict(self, recurse=True, retain_collection_types=True)
        del d["config"]
        return {k: v for k, v in d.items() if v is not None}


@attrs(slots=True, frozen=True)
class AuditContainer:
    """Container for model audit output."""

    name = attrib(type=str)
    instantiation = attrib(type=dict, validator=instance_of(dict))
    output = attrib(type=dict, validator=instance_of(dict))
    docstring = attrib(type=str, default=None, validator=optional(instance_of(str)))
    signature = attrib(type=str, default=None, validator=optional(instance_of(str)))
    steps = attrib(type=dict, default=None, validator=optional(instance_of(dict)))
    metadata = attrib(type=dict, default=None, validator=optional(instance_of(dict)))
    config = attrib(
        type=AuditConfig, factory=AuditConfig, validator=instance_of(AuditConfig)
    )

    @classmethod
    def create(cls, model, config=None):
        """Create audit"""
        if config is None:
            config = AuditConfig()
        name = model.__class__.__qualname__
        kwargs = {"name": name, "config": config}
        if config.show_docstring:
            kwargs.update({"docstring": model.__doc__})
        if config.show_signature:
            kwargs.update({"signature": f"{_make_signature(model)}"})

        kws = getfullargspec(model.__class__).kwonlyargs
        attributes = model.__model_attribute_map__
        instantiation = {attributes[kw]: deepcopy(getattr(model, kw)) for kw in kws}
        kwargs.update({"instantiation": instantiation})

        step_output, final_output = _get_model_output(model)

        if config.show_steps:
            steps = {
                step: AuditStepContainer.create(
                    getattr(model, step).func, output, config.step_config
                )
                for step, output in zip(model.__model_steps__, step_output)
            }
            kwargs.update({"steps": steps})

        kwargs.update({"output": final_output})

        return cls(**kwargs)

    def as_dict(self, include_config=True):
        d = {"name": self.name}
        if self.signature is not None:
            d.update({"signature": self.signature})
        if self.docstring is not None:
            d.update({"docstring": self.docstring})
        if self.instantiation is not None:
            d.update({"instantiation": self.instantiation})
        if self.steps is not None:
            d.update({"steps": {k: v.as_dict() for k, v in self.steps.items()}})
        if self.output is not None:
            d.update({"output": self.output})
        if self.metadata is not None:
            d.update({"metadata": self.metadata})
        if include_config is True:
            d.update({"config": asdict(self.config)})
        return d


#########################################################################################
# run_model_audit
#########################################################################################


def run_model_audit(model, file=None, **kwargs):
    config = kwargs.pop("config", None)
    if file is None:
        return AuditContainer.create(model, config=config)
    file_ext = pathlib.Path(file).suffix
    _run_model_audit(file_ext=file_ext, model=model, file=file, config=config, **kwargs)


@dispatch_function(key_parameters=("file_ext",))
def _run_model_audit(file_ext, model, file, config, **kwargs):
    """test run_model audit"""
    msg = "No registered function based on passed paramters and no default function."
    raise NotImplementedError(msg)


@_run_model_audit.register(file_ext=".xlsx")
def _(model, file, config, **kwargs):
    audit_dict = AuditContainer.create(model, config=config).as_dict()
    create_footings_xlsx_file(audit_dict, file, **kwargs)


@_run_model_audit.register(file_ext=".json")
def _(model, file, config, **kwargs):
    audit_dict = AuditContainer.create(model, config=config).as_dict()
    create_footings_json_file(audit_dict, file, **kwargs)



import pandas as pd


def convert_to_records(frame: pd.DataFrame, column_case=None):
    """Convert a DataFrame into a list of records.

    Parameters
    ----------
    frame : pd.DataFrame
        The DataFrame to convert.
    column_case : str, optional
        Set to 'lower' to convert columns to lower case and 'upper' for upper case.

    Returns
    -------
    list
        A list of records from the DataFrame.
    """
    if column_case == "upper":
        frame.columns = [col.upper() for col in frame.columns]
    elif column_case == "lower":
        frame.columns = [col.lower() for col in frame.columns]
    return frame.to_dict(orient="records")



from attr import attrib, attrs

from .data_dictionary import DataDictionary

# from attr.validators import (
#     and_,
#     deep_iterable,
#     deep_mapping,
#     in_,
#     instance_of,
#     is_callable,
#     matches_re,
#     provides,
# )


__all__ = [
    "equal_to",
    "not_equal_to",
    "isin",
]

# "greater_than",
# "greater_than_or_equal_to",
# "less_than",
# "less_than_or_equal_to",
# "in_range",
# "is_in",
# "not_in",
# "str_contains",
# "str_ends_with",
# "str_length",
# "str_matches",
# "str_starts_with",
# "deep_iterable",
# "deep_mapping",
# "is_callable",


@attrs(repr=False, slots=True, hash=True)
class _EqualToValidator:
    value = attrib()

    def _call_data_dictionary(self, inst, attr, value):
        test = value != self.value
        fails = sum(test)
        if fails > 0:
            msg = f"{attr.name} failed {repr(self)[1:-1]} {fails} out of {str(len(value))} rows."
            raise ValueError(msg)

    def __call__(self, inst, attr, value):
        if isinstance(inst, DataDictionary):
            return self._call_data_dictionary(inst, attr, value)
        if value != self.value:
            msg = f"{attr.name} value of {str(value)} does not equal {str(self.value)}."
            raise ValueError(msg)

    def __repr__(self):
        return f"<equal_to(value={str(self.value)}) validator>"


def equal_to(value):
    """A validator that raises a `ValueError` if the initializer is called with a value
    that is not equal to the value provided.

    Can be used under both a `DataDictionary` and a `Model`.

    Parameters
    ----------
    value : Any
        The value to compare for equality.

    Raises
    ------
    ValueError
        With a human readable error message, the attribute (of type `attr.Attribute`), the
        equality value, and the value passed.
    """
    return _EqualToValidator(value)


@attrs(repr=False, slots=True, hash=True)
class _NotEqualToValidator:
    value = attrib()

    def _call_data_dictionary(self, inst, attr, value):
        test = value == self.value
        fails = sum(test)
        if fails > 0:
            msg = f"{attr.name} failed {repr(self)[1:-1]} {fails} out of {str(len(value))} rows."
            raise ValueError(msg)

    def __call__(self, inst, attr, value):
        if isinstance(inst, DataDictionary):
            return self._call_data_dictionary(inst, attr, value)
        if value == self.value:
            msg = f"{attr.name} value of {str(value)} does equal {str(self.value)}."
            raise ValueError(msg)

    def __repr__(self):
        return f"<not_equal_to(value={str(self.value)}) validator>"


def not_equal_to(value):
    """A validator that raises a `ValueError` if the initializer is called with a value
    that is equal to the value provided.

    Can be used under both a `DataDictionary` and a `Model`.

    Parameters
    ----------
    value : Any
        The value to compare for equality.

    Raises
    ------
    ValueError
        With a human readable error message, the attribute (of type `attr.Attribute`), the
        equality value, and the value passed.
    """
    return _NotEqualToValidator(value)


def isin(options):
    pass


def data_dictionary_types(data_dictionary):
    pass


def data_dictionary_validators(data_dictionary):
    pass


def data_dictionary_valid(data_dictionary):
    pass



import datetime
from collections.abc import Iterable, Mapping
from typing import Any, List

import pandas as pd
from attr import asdict, attrib, attrs
from openpyxl import Workbook
from openpyxl.styles import Font, NamedStyle
from openpyxl.worksheet.worksheet import Worksheet


@attrs(slots=True, frozen=True)
class FootingsXlsxEntry:
    """FootingsXlsxEntry"""

    worksheet: str = attrib()
    source: str = attrib()
    mapping: str = attrib()
    end_point: str = attrib()
    column_name: str = attrib()
    dtype: str = attrib()
    stable: str = attrib()
    row_start: int = attrib()
    col_start: int = attrib()
    row_end: int = attrib()
    col_end: int = attrib()


def _obj_to_xlsx_cell(obj, worksheet, row, col, **kwargs):
    if isinstance(obj, pd.Timestamp):
        if all([x == 0 for x in [obj.hour, obj.minute, obj.second]]):
            obj = str(obj.date())
        else:
            obj = str(obj)
    cell = worksheet.cell(row=row, column=col)
    cell.value = obj
    if "style" in kwargs:
        cell.style = kwargs.get("style")
    if "hyperlink" in kwargs:
        cell.hyperlink = f"#{kwargs.get('hyperlink')}"


def _obj_to_xlsx_builtins(obj, worksheet, **kwargs):
    row = worksheet.row
    col = worksheet.col
    _obj_to_xlsx_cell(obj, worksheet.obj, row, col, **kwargs)
    return [
        FootingsXlsxEntry(
            worksheet=worksheet.obj.title,
            source=kwargs.get("source", None),
            mapping=kwargs.get("mapping", None),
            end_point=kwargs.get("end_point", None),
            column_name=kwargs.get("column_name", None),
            dtype=str(type(obj)),
            stable=kwargs.get("stable", None),
            row_start=row,
            col_start=col,
            row_end=row,
            col_end=col,
        )
    ]


def _obj_to_xlsx_series(obj, worksheet, **kwargs):
    row = worksheet.row
    col = worksheet.col
    _obj_to_xlsx_cell(obj.name, worksheet.obj, row, col, **kwargs)
    for idx, row_val in enumerate(obj, 1):
        _obj_to_xlsx_cell(row_val, worksheet.obj, row + idx, col, **kwargs)
    return [
        FootingsXlsxEntry(
            worksheet=worksheet.obj.title,
            source=kwargs.get("source", None),
            mapping=kwargs.get("mapping", None),
            end_point=kwargs.get("end_point", None),
            column_name=obj.name,
            dtype=str(pd.Series),
            stable=kwargs.get("stable", None),
            row_start=row,
            col_start=col,
            row_end=row + obj.shape[0],
            col_end=col,
        )
    ]


def _obj_to_xlsx_dataframe(obj, worksheet, **kwargs):
    ret = []
    start_col = worksheet.col
    for col_nm in obj:
        ret.extend(obj_to_xlsx(obj[col_nm], worksheet, **kwargs))
        worksheet.col += 1
    worksheet.col = start_col

    return ret


def _obj_to_xlsx_mapping(obj, worksheet, **kwargs):
    def _make_key(mapping, key):
        if mapping is None:
            return f"/{str(key)}/"
        return f"{mapping}{str(key)}/"

    ret = []
    mapping = kwargs.pop("mapping", None)
    for k, v in obj.items():
        if not isinstance(k, str):
            try:
                k = str(k)
            except:
                raise ValueError("Converting key of mapping to string failed.")
        kwargs.pop("end_point", None)
        key = _make_key(mapping, k)
        key_entries = obj_to_xlsx(k, worksheet, mapping=key, end_point="KEY", **kwargs)
        ret.extend(key_entries)
        worksheet.col += 1
        val_entries = obj_to_xlsx(v, worksheet, mapping=key, end_point="VALUE", **kwargs)
        ret.extend(val_entries)
        worksheet.row = max([entry.row_end for entry in val_entries]) + 1
        worksheet.col -= 1

    return ret


def _obj_to_xlsx_iterable(obj, worksheet, **kwargs):
    ret = []
    obj_len = len(obj)
    is_nested = True if isinstance(obj[0], Iterable) else False
    for idx, x in enumerate(obj):
        if idx == obj_len - 1 and is_nested:
            worksheet.row += 1
        entries = obj_to_xlsx(x, worksheet, **kwargs)
        ret.extend(entries)
        worksheet.row = max([entry.row_end for entry in entries]) + 1

    return ret


def obj_to_xlsx(obj, worksheet, **kwargs):
    """Object to xlsx"""
    builtins = [int, float, str, datetime.date, datetime.datetime, pd.Timestamp]
    if isinstance(obj, bool):
        ret = obj_to_xlsx(str(obj), worksheet, **kwargs)
    elif isinstance(obj, str) and "\n" in obj:
        new_obj = obj.split("\n")
        ret = obj_to_xlsx(new_obj, worksheet, **kwargs)
    elif any([isinstance(obj, x) for x in builtins]):
        ret = _obj_to_xlsx_builtins(obj, worksheet, **kwargs)
    elif isinstance(obj, pd.Series):
        ret = _obj_to_xlsx_series(obj, worksheet, **kwargs)
    elif isinstance(obj, pd.DataFrame):
        ret = _obj_to_xlsx_dataframe(obj, worksheet, **kwargs)
    elif isinstance(obj, Mapping):
        if len(obj) > 0:
            ret = _obj_to_xlsx_mapping(obj, worksheet, **kwargs)
        else:
            ret = _obj_to_xlsx_builtins("{}", worksheet, **kwargs)
    elif isinstance(obj, Iterable):
        if hasattr(obj, "__iter__") and not hasattr(obj, "__len__"):
            obj = list(obj)

        if len(obj) > 0:
            ret = _obj_to_xlsx_iterable(obj, worksheet, **kwargs)
        else:
            ret = _obj_to_xlsx_builtins("[]", worksheet, **kwargs)
    elif hasattr(obj, "to_audit_xlsx"):
        new_obj = obj.to_audit_xlsx()
        ret = obj_to_xlsx(new_obj, worksheet, **kwargs)
    elif callable(obj):
        new_obj = "callable: " + obj.__module__ + "." + obj.__qualname__
        ret = _obj_to_xlsx_builtins(new_obj, worksheet, **kwargs)
    elif hasattr(obj, "__str__"):
        ret = _obj_to_xlsx_builtins(str(obj), worksheet, **kwargs)
    else:
        msg = f"Not able to output {type(obj)} to an xlsx worksheet."
        raise TypeError(msg)
    return ret


#########################################################################################
# create excel file
#########################################################################################


@attrs(slots=True, repr=False)
class FootingsXlsxSheet:
    """FootingsXlsxSheet"""

    obj: Worksheet = attrib()
    row: int = attrib(default=0)
    col: int = attrib(default=0)

    def update(self, entries: List[dict], add_rows: int = 0, add_cols: int = 0):
        self.row = max([entry.row_end for entry in entries]) + add_rows
        self.col += add_cols


@attrs(slots=True, repr=False)
class FootingsXlsxWb:
    """XlsxWorkbook"""

    workbook: Workbook = attrib()
    worksheets: dict = attrib(factory=dict)
    entries: dict = attrib(factory=list)
    styles: dict = attrib(factory=dict)

    @classmethod
    def create(cls):
        """Create xlsx workbook"""
        return cls(workbook=Workbook())

    def create_sheet(
        self, name: str, start_row: int = 0, start_col: int = 0, hidden=False
    ):
        """Add worksheet"""
        wrksht = FootingsXlsxSheet(
            obj=self.workbook.create_sheet(name), row=start_row, col=start_col,
        )
        if hidden:
            wrksht.obj.sheet_state = "hidden"
        self.worksheets.update({name: wrksht})

    def add_named_style(self, name: str, style: NamedStyle):
        """Add style"""
        if isinstance(style, NamedStyle) is False:
            msg = f"The value passsed to style is not an instance of {NamedStyle}."
            raise TypeError(msg)
        self.styles.update({name: style})

    def write_obj(
        self, worksheet: str, obj: Any, add_rows: int = 0, add_cols: int = 0, **kwargs
    ):
        """Write object to worksheet"""
        wrksht = self.worksheets[worksheet]
        entries = obj_to_xlsx(obj, wrksht, **kwargs)
        wrksht.update(entries, add_rows, add_cols)
        record_entry = kwargs.get("record_entry", True)
        if record_entry:
            self.record_entry(worksheet, entries)

    def record_entry(self, worksheet: str, entries: List[dict]):
        """Record entry to __footings__ sheet."""
        for entry in entries:
            self.entries.extend([{**asdict(entry)}])

    def save(self, file: str):
        """Close workbook"""
        self.create_sheet("__footings__", 1, 1, hidden=True)
        df = pd.DataFrame.from_records(self.entries)
        self.write_obj("__footings__", df, record_entry=False)
        del self.workbook["Sheet"]
        self.workbook.save(file)


#########################################################################################
# xlsx helper functions
#########################################################################################


def _add_section(wb, sheet, section_name, section_value, source):
    wb.write_obj(
        sheet, section_name, add_cols=1, style=XLSX_FORMATS["title"], source=source
    )
    wb.write_obj(sheet, section_value, add_rows=2, add_cols=-1, source=source)


def _format_docstring(docstring):
    if docstring is None:
        return ""

    def _format_line(line, indent_len):
        if line[:indent_len] == "".join([" " for x in range(0, indent_len)]):
            return line[indent_len:]
        else:
            return line

    lines = docstring.split("\n")
    if lines[0] == "":
        lines = lines[1:]
    sections = [line for line in lines if "---" in line]
    if sections != []:
        section = sections[0]
        indent_len = len(section) - len(section.lstrip(" "))
    else:
        indent_len = 0
    return "\n".join(
        [_format_line(line, indent_len) if indent_len > 0 else line for line in lines]
    )


def _format_signature(sig):
    return sig


def _format_widths(wksht, start_column):
    for col in list(wksht.columns)[start_column:]:
        max_length = 0
        for cell in col:
            if len(str(cell.value)) > max_length:
                max_length = len(str(cell.value))
        if max_length <= 3:
            adj_width = (max_length + 1) * 1.2
        else:
            adj_width = max_length + 3
        wksht.column_dimensions[col[0].column_letter].width = adj_width


def _format_sheets(wb, sheet, format_beyond_d=False):
    wksht = wb.worksheets[sheet].obj
    wksht.sheet_view.showGridLines = False
    wksht.column_dimensions["A"].width = 2.14
    wksht.column_dimensions["B"].width = 14
    wksht.column_dimensions["C"].width = 14
    if format_beyond_d:
        _format_widths(wksht, start_column=3)


def _write_sheet(wb, sheet, step):
    _add_section(wb, sheet, "Step Name:", step["name"], "NAME")
    if "method_name" in step:
        _add_section(wb, sheet, "Method Name:", step["method_name"], "MTEHOD_NAME")
    if "docstring" in step:
        _add_section(
            wb, sheet, "Docstring:", _format_docstring(step["docstring"]), "DOCSTRING",
        )
    if "uses" in step:
        _add_section(
            wb, sheet, "Uses:", str(step["uses"]).replace("'", ""), "USES",
        )
    if "impacts" in step:
        _add_section(
            wb, sheet, "Impacts:", str(step["impacts"]).replace("'", ""), "IMPACTS",
        )
    if "metadata" in step:
        _add_section(
            wb, sheet, "Metadata:", step["metadata"], "METADATA",
        )
    if "output" in step:
        _add_section(
            wb, sheet, "Output:", step["output"], "OUTPUT",
        )


#########################################################################################
# create xlsx audit file
#########################################################################################

XLSX_FORMATS = {
    "title": NamedStyle(name="bold", font=Font(name="Calibri", bold=True)),
    "underline": NamedStyle("underline", font=Font(name="Calibri", underline="single")),
    "hyperlink": NamedStyle(
        "hyperlink", font=Font(name="Calibri", italic=True, bold=True)
    ),
}


def create_footings_xlsx_file(audit_dict, file, **kwargs):
    """Create xlsx file."""
    wb = FootingsXlsxWb.create()
    for format_nm, format_val in XLSX_FORMATS.items():
        wb.add_named_style(format_nm, format_val)

    # write main
    wb.create_sheet("Main", start_row=2, start_col=2)
    _add_section(wb, "Main", "Model Name:", audit_dict["name"], "NAME")

    if "signature" in audit_dict:
        _add_section(
            wb,
            "Main",
            "Signature:",
            _format_signature(audit_dict["signature"]),
            "SIGNATURE",
        )

    if "docstring" in audit_dict:
        _add_section(
            wb,
            "Main",
            "Docstring:",
            _format_docstring(audit_dict["docstring"]),
            "DOCSTRING",
        )

    _format_sheets(wb, "Main", format_beyond_d=False)

    # write instantiation
    wb.create_sheet("Instantiation", start_row=2, start_col=2)
    _add_section(
        wb,
        "Instantiation",
        "Instantiation:",
        audit_dict["instantiation"],
        "INSTANTIATION",
    )
    _format_widths(wb.worksheets["Instantiation"].obj, start_column=1)

    # write steps
    if "steps" in audit_dict:
        for k, v in audit_dict["steps"].items():
            name = v.get("name", None)
            wb.create_sheet(name, start_row=2, start_col=2)
            _write_sheet(wb, name, v)
            _format_sheets(wb, name, format_beyond_d=True)

    # write output
    wb.create_sheet("Output", start_row=2, start_col=2)
    _add_section(
        wb, "Output", "Output:", audit_dict["output"], "OUTPUT",
    )
    _format_sheets(wb, "Output", format_beyond_d=True)

    # save file
    wb.save(file)



import pathlib

from ..utils import dispatch_function
from .from_json import load_footings_json_file
from .from_xlsx import load_footings_xlsx_file
from .to_json import create_footings_json_file
from .to_xlsx import create_footings_xlsx_file


def load_footings_file(file: str):
    """Load footings generated file.

    Parameters
    ----------
    file : str
        The path to the file.

    Returns
    -------
    dict
        A dict representing the respective file type.

    See Also
    --------
    load_footings_json_file
    load_footings_xlsx_file
    """
    file_ext = pathlib.Path(file).suffix
    return _load_footings_file(file_ext=file_ext, file=file)


@dispatch_function(key_parameters=("file_ext",))
def _load_footings_file(file_ext, file):
    msg = f"No registered function to load a file with extension {file_ext}."
    raise NotImplementedError(msg)


@_load_footings_file.register(file_ext=".json")
def _(file):
    return load_footings_json_file(file)


@_load_footings_file.register(file_ext=".xlsx")
def _(file: str):
    return load_footings_xlsx_file(file)



import platform
from enum import Enum
from functools import partial
from typing import List, Mapping, Optional, Union

import numpy as np
import pandas as pd
from attr import attrib, attrs, evolve, make_class
from attr.validators import instance_of, optional

from .exceptions import (
    DataDictionaryPandasDtypeConversionError,
    DataDictionaryValidateError,
    DataDictionaryValidatorsConversionError,
)
from .model import def_parameter

__all__ = ["PandasDtype", "data_dictionary", "def_column"]


class PandasDtype(Enum):
    """Enum container of all valid pandas data types.

    This was copied from the `pandera` project and modified.
    """

    Bool = "bool"
    DateTime = "datetime64[ns]"
    Timedelta = "timedelta64[ns]"
    Category = "category"
    Float = "float64"
    Float16 = "float16"
    Float32 = "float32"
    Float64 = "float64"
    Int = "int32" if platform.system() == "Windows" else "int64"
    Int8 = "int8"
    Int16 = "int16"
    Int32 = "int32"
    Int64 = "int64"
    UInt8 = "uint8"
    UInt16 = "uint16"
    UInt32 = "uint32"
    UInt64 = "uint64"
    INT8 = "Int8"
    INT16 = "Int16"
    INT32 = "Int32"
    INT64 = "Int64"
    UINT8 = "UInt8"
    UINT16 = "UInt16"
    UINT32 = "UInt32"
    UINT64 = "UInt64"
    Object = "object"
    Complex = "complex128"
    Complex64 = "complex64"
    Complex128 = "complex128"
    Complex256 = "complex256"
    String = "string"


PANDAS_DTYPE_STR_ALIAS = {
    "bool": PandasDtype.Bool,
    "datetime64[ns]": PandasDtype.DateTime,
    "timedelta64[ns]": PandasDtype.Timedelta,
    "category": PandasDtype.Category,
    "float": PandasDtype.Float,
    "float16": PandasDtype.Float16,
    "float32": PandasDtype.Float32,
    "float64": PandasDtype.Float64,
    "int": PandasDtype.Int,
    "int8": PandasDtype.Int8,
    "int16": PandasDtype.Int16,
    "int32": PandasDtype.Int32,
    "int64": PandasDtype.Int64,
    "uint8": PandasDtype.UInt8,
    "uint16": PandasDtype.UInt16,
    "uint32": PandasDtype.UInt32,
    "uint64": PandasDtype.UInt64,
    "Int8": PandasDtype.INT8,
    "Int16": PandasDtype.INT16,
    "Int32": PandasDtype.INT32,
    "Int64": PandasDtype.INT64,
    "UInt8": PandasDtype.UINT8,
    "UInt16": PandasDtype.UINT16,
    "UInt32": PandasDtype.UINT32,
    "UInt64": PandasDtype.UINT64,
    "object": PandasDtype.Object,
    "complex": PandasDtype.Complex,
    "complex64": PandasDtype.Complex64,
    "complex128": PandasDtype.Complex128,
    "complex256": PandasDtype.Complex256,
    "str": PandasDtype.String,
    "string": PandasDtype.String,
}


PANDAS_DTYPE_CONVERSION = {
    PandasDtype.Bool: bool,
    PandasDtype.DateTime: pd.Timestamp,
    PandasDtype.Timedelta: pd.Timedelta,
    PandasDtype.Category: str,
    PandasDtype.Float: float,
    PandasDtype.Float16: float,
    PandasDtype.Float32: float,
    PandasDtype.Float64: float,
    PandasDtype.Int: np.int32 if platform.system() == "Windows" else np.int64,
    PandasDtype.Int8: np.int8,
    PandasDtype.Int16: np.int16,
    PandasDtype.Int32: np.int32,
    PandasDtype.Int64: np.int64,
    PandasDtype.UInt8: np.uint8,
    PandasDtype.UInt16: np.uint16,
    PandasDtype.UInt32: np.uint32,
    PandasDtype.UInt64: np.uint64,
    PandasDtype.INT8: pd.Int8Dtype,
    PandasDtype.INT16: pd.Int16Dtype,
    PandasDtype.INT32: pd.Int32Dtype,
    PandasDtype.INT64: pd.Int64Dtype,
    PandasDtype.UINT8: pd.UInt8Dtype,
    PandasDtype.UINT16: pd.UInt16Dtype,
    PandasDtype.UINT32: pd.UInt32Dtype,
    PandasDtype.UINT64: pd.UInt64Dtype,
    PandasDtype.Object: str,
    PandasDtype.Complex: complex,
    PandasDtype.Complex64: complex,
    PandasDtype.Complex128: complex,
    PandasDtype.Complex256: complex,
    PandasDtype.String: str,
}


def converter_pandas_dtype(x: Union[str, PandasDtype]):
    if x is None:
        return None
    if isinstance(x, str):
        x_upd = getattr(PandasDtype, x, None)  # try to get attribute
        if x_upd is None:  # if attribute fails lookup string alias
            x_upd = PANDAS_DTYPE_STR_ALIAS.get(x, None)
        if x_upd is None:  # if both getting attribute and string alias fail raise error.
            msg = f"The passed string [{x}] is not a built-in PandasDtype or a valid string alias."
            raise DataDictionaryPandasDtypeConversionError(msg)
        return x_upd
    elif isinstance(x, PandasDtype) is False:
        msg = "The passed object is not a built-in PandasDtype or a valid string alias."
        raise DataDictionaryPandasDtypeConversionError(msg)
    return x


def converter_validators(value: Union[str, callable]):
    def inner(x):
        if isinstance(x, str):
            msg = "The feature to look up a validator by string name is not implemented yet."
            raise NotImplementedError(msg)
        elif not callable(x):
            msg = "The passed object needs to be a callable object."
            raise DataDictionaryValidatorsConversionError(msg)
        return x

    if isinstance(value, list):
        return [inner(x) for x in value]
    return [inner(value)]


@attrs(frozen=True, slots=True)
class Column:
    """Container for column kwargs to pass into `attrs.attrib`.

    :param str name: The name of the column.
    :param Optional[Union[str, PandasDtype]] dtype: The column data type.
    :param Optional[str] description: The column description.
    :param Optional[Union[List[callable], callable]] validator: Column validators.
    :param Optional[Mapping] metadata: Optional metadata to attach to column.
    """

    name = attrib(type=str, kw_only=True, validator=instance_of(str))
    dtype = attrib(
        type=Optional[Union[str, PandasDtype]],
        default=None,
        converter=converter_pandas_dtype,
        kw_only=True,
    )
    description = attrib(
        type=Optional[str],
        default=None,
        kw_only=True,
        validator=optional(instance_of(str)),
    )
    validator = attrib(
        type=Optional[Union[List[callable], callable]],
        factory=list,
        kw_only=True,
        converter=converter_validators,
    )
    metadata = attrib(
        type=Mapping, factory=dict, kw_only=True, validator=instance_of(Mapping)
    )

    def __str__(self):
        string = self.name + " : "
        if self.dtype is not None:
            string += str(self.dtype)
        string += "\n"
        if self.description is not None:
            string += "    " + self.description
        string += "\n"
        if len(self.validator) > 0:
            for val in self.validator:
                string += str(val)
            string += "\n"
        if len(self.metadata) > 0:
            string += str(self.metadata)
            string += "\n"
        return string


def def_column(
    dtype: Optional[Union[str, PandasDtype]] = None,
    description: Optional[str] = None,
    validator: Optional[Union[List[callable], callable]] = None,
    metadata: Optional[Mapping] = None,
) -> partial:
    """Define a column with a data dictionary.

    :param Optional[Union[str, PandasDtype]] dtype: The column data type.
    :param Optional[str] description: The column description.
    :param Optional[Union[List[callable], callable]] validator: Column validators.
    :param Optional[Mapping] metadata: Optional metadata to attach to column.

    :return: A partially completed Column object (missing name which is added
        the data dictioanry is created).
    :rtype: partial
    """

    if validator is None:
        validator = []
    if metadata is None:
        metadata = {}
    return partial(
        Column,
        dtype=dtype,
        description=description,
        validator=validator,
        metadata=metadata,
    )


@attrs(frozen=True, slots=True, repr=False)
class DataDictionary:
    """The parent class of a container that provides information on a tabular data structure."""

    __columns__ = attrib(init=False, repr=False)

    @property
    def columns(self):
        """Show column names."""
        return self.__columns__

    def list_columns(self):
        """List all columns."""
        return [getattr(self, col) for col in self.__columns__]

    # def __repr__(self):
    #     name = self.__class__.__qualname__
    #     cols = "\n\t".join([repr(getattr(self, col)) for col in self.__columns__])
    #     return f"{name}[DataDictionary]\n\t{cols}\n"

    def _cols_valid(self, dataframe: pd.DataFrame):
        """Test match of columns in datadictionary vs dataframe."""
        results, msgs = [], []
        col_dict = {col.name: col for col in self.list_columns()}
        cols = list(dict.fromkeys(list(col_dict) + list(dataframe.columns)))
        for col in cols:
            dd_col = col_dict.get(col, None)
            df_col = dataframe.get(col, None)
            if dd_col is None:
                results.append(False)
                msg = f"The column [{col}] is missing from the DataDictionary but is in the dataframe."
                msgs.append(msg)
            if df_col is None:
                results.append(False)
                msg = f"The column [{col}] is missing from the dataframe but is in the DataDictionary."
                msgs.append(msg)
            if dd_col is not None and df_col is not None:
                results.append(True)
        return results, msgs

    def _types_valid(self, dataframe: pd.DataFrame):
        """Test column types in dataframe."""
        results, msgs = [], []
        for dd_col in self.list_columns():
            df_col = dataframe.get(dd_col.name, None)
            if df_col is not None:
                test_eq = dd_col.dtype.value == df_col.dtype.name
                if test_eq is False:
                    msg = f"The column [{dd_col.name}] in the DataDictionary has type [{dd_col.dtype.value}] "
                    msg += f"which is different from type in the dataframe passed [{df_col.dtype.name}]."
                    msgs.append(msg)
                results.append(test_eq)

        return results, msgs

    def _validators_valid(self, dataframe: pd.DataFrame):
        """Test validators for columns in dataframe."""
        results, msgs = [], []
        for dd_col in self.list_columns():
            df_col = dataframe.get(dd_col.name, None)
            if df_col is not None and len(dd_col.validator) > 0:
                for validator in dd_col.validator:
                    try:
                        validator(inst=self, attr=dd_col, value=df_col)
                        results.append(True)
                    except:
                        results.append(False)
                        msgs.append(
                            f"The column [{dd_col.name}] failed {str(validator)[1:-1]}."
                        )

        return results, msgs

    def validate(
        self, dataframe: pd.DataFrame, types: bool = True, validators: bool = True
    ):
        """Validate passed dataframe types and/or validators against DataDictionary.

        :param pd.DataFrame dataframe: The dataframe to validate.
        :param bool types: Test DataDictionary types against dataframe. Default is True.
        :param bool validators: Test DataDictionary validators against dataframe. Default is True.

        :return: True if test pass.
        :rtype: bool

        :raises DataDictionaryValidateError: If any of the following rules are broken -
            - All columns present in the data dictionary are in the dataframe and vice versa.
            - If types is True, test the data types of the dataframe match the data dictionary.
            - If validators is True, test that any attached validator pass for a column.
        """
        rets, msgs = self._cols_valid(dataframe)

        def _update_ret_msg(fn, dataframe):
            ret, msg = fn(dataframe)
            rets.extend(ret)
            msgs.extend(msg)

        if types is True:
            _update_ret_msg(self._types_valid, dataframe)
        if validators is True:
            _update_ret_msg(self._validators_valid, dataframe)

        if not all(rets):
            msg = "\n".join(msgs)
            raise DataDictionaryValidateError("\n" + msg)
        return True

    def def_parameter(
        self,
        column: str,
        add_dtype_validator: bool = False,
        dtype_mapping: Mapping = PANDAS_DTYPE_CONVERSION,
    ):
        """Define a parameter for a model from a data dictioanry column.

        :param str column: The column name.
        :param bool add_type_validator: If True, an instance_of validator is added to the parameter.
        :param Mapping dtype_mapping: A conversion mapping of PandasDtypes to different types.
        """
        col = getattr(self, column, None)
        if col is None:
            msg = f"The column [{column}] does not belong to the data dictionary."
            raise AttributeError(msg)
        dtype = dtype_mapping.get(col.dtype)
        validator = col.validator.copy()
        if add_dtype_validator is True:
            validator.append(instance_of(dtype))
        return def_parameter(
            dtype=dtype, description=col.description, validator=validator,
        )

    def def_sensitivity(self, column):
        raise NotImplementedError("This feature is not implemented yet.")

    def def_meta(self, column):
        raise NotImplementedError("This feature is not implemented yet.")

    def def_column(self, column):
        """Define a column for another data dictionary using an existing data dictioanry column.

        :param str column: The column name.
        """
        col = getattr(self, column, None)
        if col is None:
            msg = f"The column [{column}] does not belong to the data dictionary."
            raise AttributeError(msg)
        return col


def make_data_doc(cls_doc, cols):
    if cls_doc is None:
        doc = ""
    else:
        doc = str(cls_doc) + "\n\n"
    doc += "Columns\n"
    doc += "-------\n"
    for col in cols:
        doc += str(col)
        doc += "\n"
    return doc


def data_dictionary(cls: type = None):
    """A decorator that creates a custom data dictionary object."""
    if cls is None:
        return partial(data_dictionary)

    def make_data_dictionary(cls):
        def set_column(x):
            col = getattr(cls, x)
            if isinstance(col, Column):
                return evolve(col, name=x)
            return col(name=x)

        exclude = [x for x in DataDictionary.__dict__.keys() if x[0] != "_"]
        cols = [
            set_column(x) for x in cls.__dict__.keys() if x[0] != "_" and x not in exclude
        ]
        attrs = {col.name: attrib(default=col, init=False) for col in cols}
        orig_doc = str(cls.__doc__)
        cls = make_class(
            cls.__name__,
            attrs,
            bases=(DataDictionary,),
            slots=True,
            frozen=True,
            repr=False,
        )
        cls.__columns__ = tuple(col.name for col in cols)
        cls.__doc__ = make_data_doc(orig_doc, cols)
        return cls

    return make_data_dictionary(cls)()  # note an instance is returned



import sys
from inspect import Parameter, Signature, signature
from typing import Callable, Dict, Iterable, Optional, Tuple, Union

from attr import attrib, attrs
from attr.validators import instance_of, is_callable, optional

from .exceptions import Error

__all__ = ["create_foreach_jig"]


def _exclude_iterator_keys(iterator_keys: tuple, pass_iterator_keys: tuple):
    return set(iterator_keys).difference(pass_iterator_keys)


def _get_key(keys: tuple, **kwargs):
    if len(keys) > 1:
        iter_key = tuple(v for k, v in kwargs.items() if k in keys)
    else:
        iter_key = kwargs.get(keys[0], None)
    return iter_key if iter_key != tuple() else None


def _make_mapping_signature(iterator_keys: tuple):
    params = [Parameter(name=key, kind=Parameter.KEYWORD_ONLY) for key in iterator_keys]
    params.append(Parameter(name="kwargs", kind=Parameter.VAR_KEYWORD))

    return Signature(parameters=params, return_annotation=MappedModel)


def _make_foreach_signature(iterator_name: str, constant_params: tuple):
    params = [Parameter(name=iterator_name, kind=Parameter.KEYWORD_ONLY)]
    for param in constant_params:
        params.append(Parameter(name=param, kind=Parameter.KEYWORD_ONLY))

    return Signature(parameters=params, return_annotation=tuple)


@attrs(frozen=True)
class WrappedModel:
    """Model wrapper to catch errors when instantiating and running a model.

    :param model: The model to wrap.
    :param tuple iterator_keys: The keys identifying the record that will be passed to the model.
    :param Optional[Tuple] pass_iterator_keys: The iterator keys to pass into the model.
    :param Optional[Callable] parallel_wrap: An optional wrapper to make the model parallel (e.g., dask.delayed).
    :param  Optional[Dict] parallel_kwargs: Optional kwargs to pass to parallel_wrap.

    :return: The output of the wrapped model when calling model.run() when no
        errors occur. If an error occurs during instantiation or running the model
        an Error object is returned.
    """

    model = attrib()
    iterator_keys = attrib(type=tuple, validator=instance_of(tuple))
    pass_iterator_keys = attrib(
        type=Optional[Tuple], kw_only=True, validator=instance_of(tuple), factory=tuple,
    )
    parallel_wrap = attrib(
        type=Optional[Callable],
        default=None,
        kw_only=True,
        validator=optional(is_callable()),
    )
    parallel_kwargs = attrib(
        type=Optional[Dict],
        default=None,
        kw_only=True,
        validator=optional(instance_of(dict)),
    )
    wrapped_model = attrib(default=None, init=False, repr=False)

    def __attrs_post_init__(self):
        object.__setattr__(self, "__signature__", signature(self.model))

    def create_wrapped_model(self):
        def wrapper(**kwargs):
            try:
                excluded_keys = _exclude_iterator_keys(
                    self.iterator_keys, self.pass_iterator_keys
                )
                model_kwargs = {k: v for k, v in kwargs.items() if k not in excluded_keys}
                ret = self.model(**model_kwargs).run()
            except:
                ex_type, ex_value, ex_trace = sys.exc_info()
                key = ({k: kwargs[k] for k in self.iterator_keys},)
                ret = Error.create(key=key, sys_info=sys.exc_info())
            return ret

        if self.parallel_wrap is not None:
            if self.parallel_kwargs is None:
                wrapper = self.parallel_wrap(wrapper)
            else:
                wrapper = self.parallel_wrap(wrapper, **self.parallel_kwargs)

        object.__setattr__(self, "wrapped_model", wrapper)

    def __call__(self, **kwargs):
        if self.wrapped_model is None:
            self.create_wrapped_model()
        return self.wrapped_model(**kwargs)


@attrs(slots=True, frozen=True)
class MappedModel:
    """A mapping of models to choose from and run based on the specified iterator_keys.

    :param dict mapping: A dictonary mapping of keys to the WrappedModel to be called.
    :param tuple mapped_keys: The keys to be used to lookup the model in mapping.

    :return: The output of the wrapped models when calling model.run() when no
        errors occur. If an error occurs during instantiation or running the model
        an Error object is returned.
    """

    mapping = attrib(type=dict, validator=instance_of(dict))
    mapped_keys = attrib(type=tuple, validator=instance_of(tuple))

    @classmethod
    def create(
        cls,
        mapping: dict,
        *,
        model_wrapper: callable,
        mapped_keys: tuple,
        iterator_keys: tuple,
        pass_iterator_keys: Optional[Tuple] = None,
        parallel_wrap: Optional[Callable] = None,
        parallel_kwargs: Optional[Dict] = None,
    ):
        """Create a MappedModel.

        :param mapping: A mapping of keys to models to be called.
        :param model_wrapper: The function to be called to turn the model into a WrappedModel.
        :param tuple mapped_keys: The keys to be used to lookup the model in mapping.
        :param tuple iterator_keys: The keys identifying the record that will be passed to the model
            (passed to model_wrapper).
        :param Optional[Tuple] pass_iterator_keys: The iterator keys to pass into the model.
            (passed to model_wrapper).
        :param Optional[Callable] parallel_wrap: An optional wrapper to make the model parallel (e.g., dask.delayed).
            (passed to model_wrapper).
        :param  Optional[Dict] parallel_kwargs: Optional kwargs to pass to parallel_wrap.
            (passed to model_wrapper).

        :return: WrappedModel
        """
        kws = {
            "iterator_keys": iterator_keys,
            "pass_iterator_keys": pass_iterator_keys,
            "parallel_wrap": parallel_wrap,
            "parallel_kwargs": parallel_kwargs,
        }
        mapping = {k: model_wrapper(v, **kws) for k, v in mapping.items()}
        sig = _make_mapping_signature(iterator_keys)
        cls.__signature__ = sig

        return cls(mapped_keys=mapped_keys, mapping=mapping)

    def get_model(self, **kwargs):
        """Get model based on pass kwargs."""
        key = _get_key(self.mapped_keys, **kwargs)
        if key is None:
            msg = f"Key was not found using the passed kwargs and iterator keys of [{str(self.mapped_keys)}]."
            raise ValueError(msg)

        model = self.mapping.get(key, None)
        if model is None:
            key = {k: kwargs[k] for k in self.mapped_keys}
            msg = f"Model not found using the key of [{str(key)}]."
            raise KeyError(msg)

        return model

    def __call__(self, **kwargs):
        model = self.get_model(**kwargs)
        return model(**kwargs)


@attrs(frozen=True, slots=True)
class ForeachJig:
    """A model runs a WrappedModel or MappedModels for each item in an iterator.

    :param model: The models to call.
    :type model: Union[WrappedModel, MappedModel]
    :param str iterator_name: The name to assign the iterator to be passed (will be used in
        signature of the returned model).
    :param Optional[Tuple] constant_params: The parameter names which will be constant for all
        items in the iterator.
    :param Optional[Callable] success_wrap: An optional function to call upon running the model
        on the items that returned without error (note if none return without error an empty
        list is returned).
    :param Optional[Callable] error_wrap: An optional function to call upon running the model
        on the items that returned with error (note if none return with error an empty list is
        returned).
    :param Optional[Callable] compute: An optional function to be used to call compute or
        return on the modeled objects. This is to be paired with parallel tools such as
        dask.compute or ray.get.
    :param Optional[Dict] compute_kwargs: Optional kwargs to pass into compute.
    """

    model = attrib(
        type=Union[WrappedModel, MappedModel],
        validator=instance_of((WrappedModel, MappedModel,)),
    )
    iterator_name = attrib(type=str, validator=instance_of(str))
    constant_params = attrib(type=Optional[Tuple], validator=instance_of(tuple))
    success_wrap = attrib(type=Optional[Callable], validator=optional(is_callable()))
    error_wrap = attrib(type=Optional[Callable], validator=optional(is_callable()))
    compute = attrib(type=Optional[Callable], validator=optional(is_callable()))
    compute_kwargs = attrib(type=Optional[Dict], validator=instance_of(dict))

    @classmethod
    def create(
        cls,
        model: Union[WrappedModel, MappedModel],
        iterator_name: str,
        constant_params: Optional[Tuple] = None,
        success_wrap: Optional[Callable] = None,
        error_wrap: Optional[Callable] = None,
        compute: Optional[Callable] = None,
        compute_kwargs: Optional[Dict] = None,
    ):
        """A model runs a WrappedModel or MappedModels for each item in an iterator.

        :param model: The models to call.
        :type model: Union[WrappedModel, MappedModel]
        :param str iterator_name: The name to assign the iterator to be passed (will be used in
            signature of the returned model).
        :param Optional[Tuple] constant_params: The parameter names which will be constant for all
            items in the iterator.
        :param Optional[Callable] success_wrap: An optional function to call upon running the model
            on the items that returned without error (note if none return without error an empty
            list is returned).
        :param Optional[Callable] error_wrap: An optional function to call upon running the model
            on the items that returned with error (note if none return with error an empty list is
            returned).
        :param Optional[Callable] compute: An optional function to be used to call compute or
            return on the modeled objects. This is to be paired with parallel tools such as
            dask.compute or ray.get.
        :param Optional[Dict] compute_kwargs: Optional kwargs to pass into compute.
        """

        if constant_params is None:
            constant_params = tuple()

        if compute_kwargs is None:
            compute_kwargs = {}

        cls.__signature__ = _make_foreach_signature(iterator_name, constant_params)
        return cls(
            model=model,
            iterator_name=iterator_name,
            constant_params=constant_params,
            success_wrap=success_wrap,
            error_wrap=error_wrap,
            compute=compute,
            compute_kwargs=compute_kwargs,
        )

    def __call__(self, **kwargs):
        """Calls the underlying WrappedModel or MappedModel for each item in the named iterator.

        :return: A tuple where the first item are those items that have been successfully ran through
            the underlying models and the second item are those items that failed.
        """
        iterator = kwargs.pop(self.iterator_name)
        if not isinstance(iterator, Iterable):
            raise TypeError("The specified iterator object is not an iterator.")
        output = [self.model(**entry, **kwargs) for entry in iterator]
        if self.compute is not None:
            output = self.compute(output, **self.compute_kwargs)
        successes, errors = [], []
        for result in output:
            (errors if isinstance(result, Error) else successes).append(result)
        if self.success_wrap is not None and len(successes) > 0:
            successes = self.success_wrap(successes)
        if self.error_wrap is not None and len(errors) > 0:
            errors = self.error_wrap(errors)
        return (successes, errors)


def create_foreach_jig(
    model,
    *,
    iterator_name: str,
    iterator_keys: tuple,
    mapped_keys: Optional[Tuple] = None,
    constant_params: Optional[Tuple] = None,
    pass_iterator_keys: Optional[Tuple] = None,
    success_wrap: Optional[Callable] = None,
    error_wrap: Optional[Callable] = None,
):
    """Create a ForeachJig that runs a WrappedModel or MappedModels for each item in an iterator.

    :param model: The models to call.
    :type model: Union[WrappedModel, MappedModel]
    :param str iterator_name: The name to assign the iterator to be passed (will be used in
        signature of the returned model).
    :param Optional[Tuple] mapped_keys: The keys to be used to lookup the model in mapping.
    :param Optional[Tuple] constant_params: The parameter names which will be constant for all
        items in the iterator.
    :param Optional[Callable] success_wrap: An optional function to call upon running the model
        on the items that returned without error (note if none return without error an empty
        list is returned).
    :param Optional[Callable] error_wrap: An optional function to call upon running the model
        on the items that returned with error (note if none return with error an empty list is
        returned).
    :param Optional[Callable] compute: An optional function to be used to call compute or
        return on the modeled objects. This is to be paired with parallel tools such as
        dask.compute or ray.get.
    :param Optional[Dict] compute_kwargs: Optional kwargs to pass into compute.

    :return: ForeachJig (with updated signature)
    """
    if isinstance(model, dict):
        model = MappedModel.create(
            model,
            iterator_keys=iterator_keys,
            mapped_keys=mapped_keys,
            pass_iterator_keys=pass_iterator_keys,
        )
    else:
        model = WrappedModel(
            model, iterator_keys=iterator_keys, pass_iterator_keys=pass_iterator_keys
        )

    return ForeachJig.create(
        model=model,
        iterator_name=iterator_name,
        constant_params=constant_params,
        success_wrap=success_wrap,
        error_wrap=error_wrap,
    )



from datetime import datetime

from attr import Factory


def _run_date_time(strftime=None):
    """Run date time"""
    if strftime is not None:
        return datetime.now().strftime(strftime)
    return datetime.now()


run_date_time = Factory(_run_date_time)



from typing import Callable, Dict, Optional, Tuple

from dask import compute, delayed

from ..jigs import ForeachJig, MappedModel, WrappedModel

__all__ = ["create_dask_foreach_jig"]


def compute_wrapper(output, **compute_kwargs):
    return compute(output, **compute_kwargs)[0]


def create_dask_foreach_jig(
    model,
    *,
    iterator_name: str,
    iterator_keys: tuple,
    mapped_keys: Optional[Tuple] = None,
    constant_params: Optional[Tuple] = None,
    pass_iterator_keys: Optional[Tuple] = None,
    success_wrap: Optional[Callable] = None,
    error_wrap: Optional[Callable] = None,
    dask_delayed_kwargs: Optional[Dict] = None,
    dask_compute_kwargs: Optional[Dict] = None,
):
    """Create a dask backed ForeachJig that runs a WrappedModel or MappedModels for each item in an iterator.

    :param model: The models to call.
    :type model: Union[WrappedModel, MappedModel]
    :param str iterator_name: The name to assign the iterator to be passed (will be used in
        signature of the returned model).
    :param Optional[Tuple] mapped_keys: The keys to be used to lookup the model in mapping.
    :param Optional[Tuple] constant_params: The parameter names which will be constant for all
        items in the iterator.
    :param Optional[Callable] success_wrap: An optional function to call upon running the model
        on the items that returned without error (note if none return without error an empty
        list is returned).
    :param Optional[Callable] error_wrap: An optional function to call upon running the model
        on the items that returned with error (note if none return with error an empty list is
        returned).
    :param Optional[Dict] dask_delayed_kwargs: Optional kwargs to pass into dask.dealyed.
    :param Optional[Dict] dask_compute_kwargs: Optional kwargs to pass into dask.compute.

    :return: ForeachJig (with updated signature)
    """

    if isinstance(model, dict):
        if mapped_keys is None:
            msg = (
                "When passing a dict of models, the keys used must be set in mapped_keys."
            )
            raise ValueError(msg)
        model = MappedModel.create(
            model,
            model_wrapper=WrappedModel,
            iterator_keys=iterator_keys,
            mapped_keys=mapped_keys,
            pass_iterator_keys=pass_iterator_keys,
            parallel_wrap=delayed,
            parallel_kwargs=dask_delayed_kwargs,
        )
    else:
        model = WrappedModel(
            model,
            iterator_keys=iterator_keys,
            pass_iterator_keys=pass_iterator_keys,
            parallel_wrap=delayed,
            parallel_kwargs=dask_delayed_kwargs,
        )

    return ForeachJig.create(
        model=model,
        iterator_name=iterator_name,
        constant_params=constant_params,
        success_wrap=success_wrap,
        error_wrap=error_wrap,
        compute=compute_wrapper,
        compute_kwargs=dask_compute_kwargs,
    )



import sys
from enum import Enum, auto
from functools import partial
from traceback import extract_tb, format_list
from typing import Any, List, Optional

import numpydoc.docscrape as numpydoc
from attr import NOTHING, attrib, attrs, evolve, make_class
from attr._make import _CountingAttr
from attr.setters import NO_OP, frozen

from .audit import run_model_audit
from .doc_tools.docscrape import FootingsDoc
from .exceptions import ModelCreationError, ModelRunError
from .visualize import visualize_model

__all__ = [
    "Model",
    "model",
    "step",
    "def_parameter",
    "def_sensitivity",
    "def_meta",
    "def_intermediate",
    "def_return",
]


class ModelAttributeType(Enum):
    """Allowed dodel attribute types belonging to a model."""

    Parameter = auto()
    Sensitivity = auto()
    Meta = auto()
    Intermediate = auto()
    Return = auto()


def _define(
    attribute_type: ModelAttributeType,
    init: bool,
    frozen: bool,
    dtype: Optional[Any] = None,
    description: Optional[str] = None,
    default: Optional[Any] = NOTHING,
    validator: Optional[callable] = None,
    converter: Optional[callable] = None,
    **kwargs,
):
    metadata = {
        "description": description if description is not None else "",
        "footings_attribute_type": attribute_type,
    }
    on_setattr = NO_OP if frozen is False else None

    return attrib(
        init=init,
        type=dtype,
        repr=False,
        default=default,
        validator=validator,
        converter=converter,
        metadata=metadata,
        on_setattr=on_setattr,
        kw_only=True,
        **kwargs,
    )


def def_parameter(
    *,
    dtype: Optional[Any] = None,
    description: Optional[str] = None,
    converter: Optional[callable] = None,
    validator: Optional[callable] = None,
    **kwargs,
):
    """Define a parameter attribute under the model.

    A parameter attribute is a frozen attribute that is passed on instantiation of the model.

    :param Optional[Any] dtype: The expected type of the attribute. If not None, value will be
        validated on instantiation.
    :param Optional[str] description: Optional description to add.
    :param Optional[callable] converter: Optional callable that is used to convert value to desired format.
    :param Optional[callable] validator: Optional callaable that is used to validate value.
    :param kwargs: Advanced options to pass through to `attrs.attrib`.
    """
    return _define(
        attribute_type=ModelAttributeType.Parameter,
        init=True,
        dtype=dtype,
        description=description,
        converter=converter,
        validator=validator,
        frozen=True,
        **kwargs,
    )


def def_sensitivity(
    *,
    default: Any = NOTHING,
    dtype: Optional[Any] = None,
    description: Optional[str] = None,
    converter: Optional[callable] = None,
    validator: Optional[callable] = None,
    **kwargs,
):
    """Define a sensitivity attribute under the model.

    A sensitivity attribute is a frozen attribute with a default value that is passed on
    instantiation of the model.

    :param Any default: The default value of the sensitivity.
    :param Optional[Any] dtype: The expected type of the attribute. If not None, value will be
        validated on instantiation.
    :param Optional[str] description: Optional description to add.
    :param Optional[callable] converter: Optional callable that is used to convert value to desired format.
    :param Optional[callable] validator: Optional callaable that is used to validate value.
    :param kwargs: Advanced options to pass through to `attrs.attrib`.
    """
    return _define(
        attribute_type=ModelAttributeType.Sensitivity,
        init=True,
        dtype=dtype,
        description=description,
        default=default,
        converter=converter,
        validator=validator,
        frozen=True,
        **kwargs,
    )


def def_meta(
    *,
    meta: Any,
    dtype: Optional[Any] = None,
    description: Optional[str] = None,
    converter: Optional[callable] = None,
    validator: Optional[callable] = None,
    **kwargs,
):
    """Define a meta attribute under the model.

    A meta attribute is a frozen attribute that is passed on instantiation of the model.

    :param Any meta: The meta value to pass to the model.
    :param Optional[Any] dtype: The expected type of the attribute. If not None, value will be
        validated on instantiation.
    :param Optional[str] description: Optional description to add.
    :param Optional[callable] converter: Optional callable that is used to convert value to desired format.
    :param Optional[callable] validator: Optional callaable that is used to validate value.
    :param kwargs: Advanced options to pass through to `attrs.attrib`.
    """
    return _define(
        attribute_type=ModelAttributeType.Meta,
        init=False,
        dtype=dtype,
        description=description,
        default=meta,
        converter=converter,
        validator=validator,
        frozen=True,
        **kwargs,
    )


def def_intermediate(
    *,
    dtype: Optional[Any] = None,
    description: Optional[str] = None,
    init_value: Optional[Any] = None,
    **kwargs,
):
    """Define an intermediate attribute under the model.

    A placeholder is a non-frozen attribute that is created by the model and not returned when
    the model runs. It can be used to hold intermediate values for calculation.

    :param Optional[Any] dtype: The expected type of the attribute. If not None, value will be
        validated on instantiation.
    :param Optional[str] description: Optional description to add.
    :param Optional[Any] init_value: Optional initival value to assign.
    :param kwargs: Advanced options to pass through to `attrs.attrib`.
    """
    return _define(
        attribute_type=ModelAttributeType.Intermediate,
        init=False,
        dtype=dtype,
        description=description,
        default=init_value,
        frozen=False,
        **kwargs,
    )


def def_return(
    *,
    dtype: Optional[Any] = None,
    description: Optional[str] = None,
    init_value: Optional[Any] = None,
    **kwargs,
):
    """Define an intermediate attribute under the model.

    A placeholder is a non-frozen attribute that is created by the model and returned when
    the model runs.

    :param Optional[Any] dtype: The expected type of the attribute. If not None, value will be
        validated on instantiation.
    :param Optional[str] description: Optional description to add.
    :param Optional[Any] init_value: Optional initival value to assign.
    :param kwargs: Advanced options to pass through to `attrs.attrib`.
    """
    return _define(
        attribute_type=ModelAttributeType.Return,
        init=False,
        dtype=dtype,
        description=description,
        default=init_value,
        frozen=False,
        **kwargs,
    )


def _run(self, to_step):
    if len(self.__model_steps__) == 0:
        raise ModelRunError("Not able to run model because no steps are registered.")
    if len(self.__model_returns__) == 0:
        raise ModelRunError(
            "Not able to run model because no return attributes are registered."
        )
    if to_step is None:
        steps = self.__model_steps__
    else:
        try:
            position = self.__model_steps__.index(to_step)
            steps = self.__model_steps__[: (position + 1)]
        except ValueError as e:
            msg = f"The step passed to to_step '{to_step}' does not exist as a step."
            raise e(msg)

    def _run_step(step):
        try:
            return getattr(self, step)()
        except:
            exc_type, exc_value, exc_trace = sys.exc_info()
            msg = f"At step [{step}], an error occured.\n"
            msg += f"  Error Type = {exc_type.__name__}\n"
            msg += f"  Error Message = {exc_value}\n"
            msg += f"  Error Trace = {format_list(extract_tb(exc_trace))}\n"
            raise ModelRunError(msg)

    for step in steps:
        _run_step(step)

    if to_step is not None:
        return self
    if len(self.__model_returns__) > 1:
        return tuple(getattr(self, ret) for ret in self.__model_returns__)
    return getattr(self, self.__model_returns__[0])


@attrs(slots=True, repr=False)
class Model:
    """The parent modeling class providing the key methods of run, audit, and visualize."""

    __model_steps__: tuple = attrib(init=False, repr=False)
    __model_parameters__: tuple = attrib(init=False, repr=False)
    __model_sensitivities__: tuple = attrib(init=False, repr=False)
    __model_meta__: tuple = attrib(init=False, repr=False)
    __model_intermediates__: tuple = attrib(init=False, repr=False)
    __model_returns__: tuple = attrib(init=False, repr=False)
    __model_attribute_map__: dict = attrib(init=False, repr=False)

    def visualize(self):
        """Visualize the model to get an understanding of what model attributes are used and when."""
        return visualize_model(self)

    def audit(self, file: str = None, **kwargs):
        """Audit the model which returns copies of the object as it is modified across each step.

        Parameters
        ----------
        file : str, optional
            The name of the audit output file.
        kwargs
            Additional key words passed to audit.

        Returns
        -------
        None
            An audit file in specfified format (e.g., .xlsx).
        """
        return run_model_audit(model=self, file=file, **kwargs)

    def run(self, to_step=None):
        """Runs the model and returns any returns defined.

        Parameters
        ----------
        to_step : str, optional
            The name of the step to run model to.

        """
        return _run(self, to_step=to_step)


@attrs(frozen=True, slots=True)
class Step:
    name = attrib(type=str)
    docstring = attrib(type=str)
    method = attrib(type=callable)
    method_name = attrib(type=str)
    uses = attrib(type=List[str])
    impacts = attrib(type=List[str])
    metadata = attrib(type=dict, factory=dict)

    @classmethod
    def create(
        cls,
        method: callable,
        uses: List[str],
        impacts: List[str],
        name: Optional[str] = None,
        docstring: Optional[str] = None,
        metadata: Optional[dict] = None,
    ):
        method_name = method.__qualname__.split(".")[1]
        return cls(
            name=name if name is not None else method_name,
            docstring=docstring if docstring is not None else method.__doc__,
            method=method,
            method_name=method_name,
            uses=uses,
            impacts=impacts,
            metadata={} if metadata is None else metadata,
        )

    def __doc__(self):
        return self.docstring

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        return partial(self, obj)

    def __call__(self, *args, **kwargs):
        return self.method(*args, **kwargs)


def step(
    method: callable = None,
    *,
    uses: List[str],
    impacts: List[str],
    name: str = None,
    docstring: str = None,
    metadata: dict = None,
):
    """Turn a method into a step within the footings framework.

    Parameters
    ----------
    method : callable, optional
        The method to decorate, by default None.
    uses : List[str]
        A list of the object names used by the step.
    impacts : List[str]
        A list of the object names that are impacted by the step (i.e., the returns and intermediates).
    wrap : callable, optional
        Wrap or source the docstring from another object, by default None.

    Returns
    -------
    callable
        The decorated method with a attributes for uses and impacts and updated docstring if wrap passed.
    """
    if method is None:
        return partial(
            step,
            uses=uses,
            impacts=impacts,
            name=name,
            docstring=docstring,
            metadata=metadata,
        )

    return Step.create(
        method=method,
        uses=uses,
        impacts=impacts,
        name=name,
        docstring=docstring,
        metadata=metadata,
    )


def _make_doc_parameter(attribute):
    if attribute.type is None:
        atype = ""
    elif isinstance(attribute.type, str):
        atype = attribute.type
    elif hasattr(attribute.type, "__qualname__"):
        atype = attribute.type.__qualname__
    else:
        try:
            atype = str(attribute.type)
        except:
            atype = ""
    return numpydoc.Parameter(
        attribute.name, atype, [attribute.metadata.get("description", "")]
    )


def _parse_attriubtes(cls):
    sections = ["Parameters", "Sensitivities", "Meta", "Intermediates", "Returns"]
    parsed_attributes = {section: [] for section in sections}

    for attribute in cls.__attrs_attrs__:
        attribute_type = attribute.metadata.get("footings_attribute_type", None)
        if attribute_type is ModelAttributeType.Parameter:
            parsed_attributes["Parameters"].append(_make_doc_parameter(attribute))
        elif attribute_type is ModelAttributeType.Sensitivity:
            parsed_attributes["Sensitivities"].append(_make_doc_parameter(attribute))
        elif attribute_type is ModelAttributeType.Meta:
            parsed_attributes["Meta"].append(_make_doc_parameter(attribute))
        elif attribute_type is ModelAttributeType.Intermediate:
            parsed_attributes["Intermediates"].append(_make_doc_parameter(attribute))
        elif attribute_type is ModelAttributeType.Return:
            parsed_attributes["Returns"].append(_make_doc_parameter(attribute))

    return parsed_attributes


def _generate_steps_sections(cls, steps):
    return [
        f"{idx}) {getattr(cls, step).name} - {getattr(cls, step).docstring}"
        for idx, step in enumerate(steps, start=1)
    ]


def _attr_doc(cls, steps):

    parsed_attributes = _parse_attriubtes(cls)
    doc = FootingsDoc(cls)

    doc["Parameters"] = parsed_attributes["Parameters"]
    doc["Sensitivities"] = parsed_attributes["Sensitivities"]
    doc["Meta"] = parsed_attributes["Meta"]
    doc["Intermediates"] = parsed_attributes["Intermediates"]
    doc["Returns"] = parsed_attributes["Returns"]
    doc["Steps"] = _generate_steps_sections(cls, steps)
    doc["Methods"] = []

    cls.__doc__ = str(doc)
    return cls


def _update_uses_impacts(src, attribute_map):
    flags = ["parameter.", "sensitivity.", "meta.", "intermediate.", "return."]

    def inner(x):
        if any([flag in x for flag in flags]):
            return x
        if x not in attribute_map:
            msg = f"The attribute [{x}] does not belong to the model."
            raise ModelCreationError(msg)
        return attribute_map[x]

    return tuple([inner(x) for x in src])


def model(cls: type = None, *, steps: List[str] = []):
    """Turn a class into a model within the footings framework.

    Parameters
    ----------
    cls : type
        The class to turn into a model.
    steps : List[str], optional
        The list of steps to the model.

    Returns
    -------
    cls
        Returns cls as a model within footings framework.
    """
    if cls is None:
        return partial(model, steps=steps)

    def inner(cls):
        # In order to be instantiated as a model, need to pass the following test.

        # 1. All attributes need to belong to a footings_attribute_type
        exclude = ["run", "audit", "visualize"]
        attributes = [x for x in cls.__dict__.keys() if x[0] != "_" and x not in exclude]
        if hasattr(cls, "__attrs_attrs__"):
            attributes += [x.name for x in cls.__attrs_attrs__ if x.name[0] != "_"]
            attrs_attrs = {x.name: x for x in cls.__attrs_attrs__ if x.name[0] != "_"}
        else:
            attrs_attrs = {}
        parameters, sensitivities, meta, intermediates, returns = [], [], [], [], []
        for attribute in attributes:
            attr = getattr(cls, attribute)
            if attribute in attrs_attrs:
                attr = attrs_attrs[attribute]
            else:
                if isinstance(attr, _CountingAttr) is False:
                    msg = f"The attribute {attribute} is not registered to a known Models group.\n"
                    msg += "Use one of def_* functions from the footings library when building a model."
                    raise ModelCreationError(msg)
            attribute_type = attr.metadata.get("footings_attribute_type", None)
            if (
                attribute_type is None
                or isinstance(attribute_type, ModelAttributeType) is False
            ):
                msg = f"The attribute {attribute} is not registered to a known Models attribute type.\n"
                msg += "Use one of def_* functions from the footings library when building a model."
                raise ModelCreationError(msg)
            if attribute_type is ModelAttributeType.Parameter:
                parameters.append(attribute)
            elif attribute_type is ModelAttributeType.Sensitivity:
                sensitivities.append(attribute)
            elif attribute_type is ModelAttributeType.Meta:
                meta.append(attribute)
            elif attribute_type is ModelAttributeType.Intermediate:
                intermediates.append(attribute)
            elif attribute_type is ModelAttributeType.Return:
                returns.append(attribute)

        # 2. For steps -
        #    - all steps are methods of cls
        #    - all steps have attributes uses and impacts
        missing_steps = []
        missing_attributes = []
        for step_nm in steps:
            step = getattr(cls, step_nm, None)
            if step is None:
                missing_steps.append(step_nm)
            if hasattr(step, "uses") is False or hasattr(step, "impacts") is False:
                missing_attributes.append(step_nm)
        if len(missing_steps) > 0:
            raise ModelCreationError(
                f"The following steps listed are missing - {str(missing_steps)} from the object."
            )
        if len(missing_attributes) > 0:
            raise ModelCreationError(
                f"The followings steps listed do not appear to be decorated steps - {str(missing_attributes)}."
            )

        # If all test pass, update steps and returns with known values as defaults.
        cls.__model_steps__ = tuple(steps)
        cls.__model_parameters__ = tuple(parameters)
        cls.__model_sensitivities__ = tuple(sensitivities)
        cls.__model_meta__ = tuple(meta)
        cls.__model_intermediates__ = tuple(intermediates)
        cls.__model_returns__ = tuple(returns)

        attribute_map = {
            **{p: f"parameter.{p}" for p in parameters},
            **{m: f"sensitivity.{m}" for m in sensitivities},
            **{m: f"meta.{m}" for m in meta},
            **{p: f"intermediate.{p}" for p in intermediates},
            **{a: f"return.{a}" for a in returns},
        }

        for step in steps:
            use_old = getattr(cls, step).uses
            use_new = _update_uses_impacts(use_old, attribute_map)
            impact_old = getattr(cls, step).impacts
            impact_new = _update_uses_impacts(impact_old, attribute_map)
            new_step = evolve(getattr(cls, step), uses=use_new, impacts=tuple(impact_new))
            setattr(cls, step, new_step)
        cls.__model_attribute_map__ = attribute_map

        exclude = [x for x in Model.__dict__.keys() if x[0] != "_"]
        attrs = {
            x: getattr(cls, x)
            for x in cls.__dict__.keys()
            if x[0] != "_" and x not in exclude
        }

        # Make attrs dataclass and update signature
        cls = make_class(
            cls.__name__,
            attrs=attrs,
            bases=(cls, Model,),
            kw_only=True,
            on_setattr=frozen,
            repr=False,
            slots=True,
        )
        return _attr_doc(cls, steps)

    return inner(cls)



from numpydoc.docscrape import ClassDoc, Reader


class FootingsDoc(ClassDoc):
    sections = {
        "Signature": "",
        "Summary": [""],
        "Extended Summary": [],
        "Parameters": [],
        "Sensitivities": [],
        "Meta": [],
        "Intermediates": [],
        "Returns": [],
        "Yields": [],
        "Receives": [],
        "Steps": [],
        "Raises": [],
        "Warns": [],
        "Other Parameters": [],
        "Attributes": [],
        "Methods": [],
        "See Also": [],
        "Notes": [],
        "Warnings": [],
        "References": "",
        "Examples": "",
        "index": {},
    }

    def _parse(self):
        self._doc.reset()
        self._parse_summary()

        sections = list(self._read_sections())
        section_names = set([section for section, content in sections])

        has_returns = "Returns" in section_names
        has_yields = "Yields" in section_names
        # We could do more tests, but we are not. Arbitrarily.
        if has_returns and has_yields:
            msg = "Docstring contains both a Returns and Yields section."
            raise ValueError(msg)
        if not has_yields and "Receives" in section_names:
            msg = "Docstring contains a Receives section but not Yields."
            raise ValueError(msg)

        for (section, content) in sections:
            if not section.startswith(".."):
                section = (s.capitalize() for s in section.split(" "))
                section = " ".join(section)
                if self.get(section):
                    self._error_location(
                        "The section %s appears twice in  %s"
                        % (section, "\n".join(self._doc._str))
                    )

            if section in (
                "Parameters",
                "Sensitivities",
                "Meta",
                "Intermediates",
                "Returns",
                "Other Parameters",
                "Attributes",
                "Methods",
            ):
                self[section] = self._parse_param_list(content)
            elif section == "Steps":
                self[section] = self._parse_steps(content)
            elif section in ("Raises", "Warns", "Receives"):
                self[section] = self._parse_param_list(
                    content, single_element_is_type=True
                )
            elif section.startswith(".. index::"):
                self["index"] = self._parse_index(section, content)
            elif section == "See Also":
                self["See Also"] = self._parse_see_also(content)
            else:
                self[section] = content

    def _parse_steps(self, content):
        r = Reader(content)
        list_ = []
        while not r.eof():
            line = r.read().strip()
            list_.append(line)
        return list_

    def _str_steps(self, name):
        out = []
        if self[name]:
            out += self._str_header(name)
            out += self[name]
            out += [""]
        return out

    def __str__(self, func_role=""):
        out = []
        out += self._str_signature()
        out += self._str_summary()
        out += self._str_extended_summary()
        for param_list in (
            "Parameters",
            "Sensitivities",
            "Meta",
            "Intermediates",
            "Returns",
            "Yields",
            "Receives",
            "Other Parameters",
        ):
            out += self._str_param_list(param_list)
        out += self._str_steps("Steps")
        for param_list in (
            "Raises",
            "Warns",
        ):
            out += self._str_param_list(param_list)
        out += self._str_section("Warnings")
        out += self._str_see_also(func_role)
        for s in ("Notes", "References", "Examples"):
            out += self._str_section(s)
        for param_list in ("Methods",):
            out += self._str_param_list(param_list)
        out += self._str_index()
        return "\n".join(out)



from typing import Callable, Dict, Optional, Tuple

# import ray
#
# from ..jigs import WrappedModel, MappedModel, ForeachJig
#
# __all__ = ["create_ray_foreach_jig"]
#
#
# def ray_wrapper(func, **kwargs):
#     def inner(func, **kwargs):
#         return ray.remote(func, **kwargs)
#
#     return inner(func, **kwargs).remote


def create_ray_foreach_jig(
    model,
    *,
    iterator_name: str,
    iterator_keys: tuple,
    mapped_keys: Optional[Tuple] = None,
    constant_params: Optional[Tuple] = None,
    pass_iterator_keys: Optional[Tuple] = None,
    success_wrap: Optional[Callable] = None,
    error_wrap: Optional[Callable] = None,
    ray_remote_kwargs: Optional[Dict] = None,
    ray_get_kwargs: Optional[Dict] = None,
):
    """Create a ray backed ForeachJig that runs a WrappedModel or MappedModels for each item in an iterator.

    :param model: The models to call.
    :type model: Union[WrappedModel, MappedModel]
    :param str iterator_name: The name to assign the iterator to be passed (will be used in
        signature of the returned model).
    :param Optional[Tuple] mapped_keys: The keys to be used to lookup the model in mapping.
    :param Optional[Tuple] constant_params: The parameter names which will be constant for all
        items in the iterator.
    :param Optional[Callable] success_wrap: An optional function to call upon running the model
        on the items that returned without error (note if none return without error an empty
        list is returned).
    :param Optional[Callable] error_wrap: An optional function to call upon running the model
        on the items that returned with error (note if none return with error an empty list is
        returned).
    :param Optional[Dict] ray_remote_kwargs: Optional kwargs to pass into ray.remote.
    :param Optional[Dict] ray_get_kwargs: Optional kwargs to pass into ray.get.

    :return: ForeachJig (with updated signature)
    """
    pass


#     if isinstance(model, dict):
#         if mapped_keys is None:
#             msg = (
#                 "When passing a dict of models, the keys used must be set in mapped_keys."
#             )
#             raise ValueError(msg)
#         model = MappedModel.create(
#             model,
#             model_wrapper=WrappedModel,
#             iterator_keys=iterator_keys,
#             mapped_keys=mapped_keys,
#             pass_iterator_keys=pass_iterator_keys,
#             parallel_wrap=ray_wrapper,
#             parallel_kwargs=ray_remote_kwargs,
#         )
#     else:
#         model = WrappedModel(
#             model,
#             iterator_keys=iterator_keys,
#             pass_iterator_keys=pass_iterator_keys,
#             parallel_wrap=ray_wrapper,
#             parallel_kwargs=ray_remote_kwargs,
#         )
#
#     return ForeachJig.create(
#         model=model,
#         iterator_name=iterator_name,
#         constant_params=constant_params,
#         success_wrap=success_wrap,
#         error_wrap=error_wrap,
#         compute=ray.get,
#         compute_kwargs=ray_get_kwargs,
#     )
#



import json
import math
from datetime import date, datetime
from inspect import isclass

import numpy as np
import pandas as pd

# from ..exceptions import Error


def _set_key(k):
    if type(k) not in [str, int, float, bool, None]:
        return str(k)
    return k


def _column_to_list(col):
    def _set_val(val):
        if isinstance(val, (int, float, bool)):
            if math.isnan(val):
                return None
            else:
                return val
        return val

    col = col.to_list()
    if any(math.isnan(val) for val in col if isinstance(val, (int, float, bool))):
        col = [_set_val(val) for val in col]
    return col


class AuditJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if hasattr(obj, "to_audit_json"):
            return obj.to_audit_json()
        elif isinstance(obj, (int, float, bool)):
            if math.isnan(obj):
                return None
            return obj
        elif isinstance(obj, np.generic):
            return obj.item()
        elif isinstance(obj, pd.DataFrame):
            return {col: _column_to_list(obj[col]) for col in obj.columns}
        elif isinstance(obj, pd.Series):
            return {obj.name: _column_to_list(obj)}
        elif isinstance(obj, (pd.Timestamp, date, datetime)):
            return str(obj)
        # elif isinstance(obj, Error):
        #     return obj.to_audit_json()
        elif callable(obj):
            return "callable: " + obj.__module__ + "." + obj.__qualname__
        elif isclass(obj) and issubclass(obj, Exception):
            return obj.__name__
        else:
            return super().default(obj)


def create_footings_json_file(d, file, **kwargs):
    with open(file, "w") as data_file:
        json.dump(obj=d, fp=data_file, cls=AuditJSONEncoder, indent=2, **kwargs)



import pandas as pd
from openpyxl import load_workbook

from ..io.to_xlsx import FootingsXlsxEntry


def load_footings_xlsx_file(file: str):
    """Load footings generated xlsx file.

    Parameters
    ----------
    file : str
        The path to the file.

    Returns
    -------
    dict
        A dict representing the respective file type.
    """

    def _make_key(x: pd.Series):
        return FootingsXlsxEntry(
            worksheet=x.worksheet,
            source=x.source,
            mapping=x.mapping,
            end_point=x.end_point,
            column_name=x.column_name,
            dtype=x.dtype,
            stable=x.stable,
            row_start=x.row_start,
            col_start=x.col_start,
            row_end=x.row_end,
            col_end=x.col_end,
        )

    wb = load_workbook(file)
    expected_cols = set([x for x in dir(FootingsXlsxEntry) if x[0] != "_"])

    if "__footings__" not in wb.sheetnames:
        msg = (
            "The workbook is missing the sheet __footings__ which holds key information."
        )
        raise ValueError(msg)

    ws = wb["__footings__"]
    data = ws.values
    cols = next(data)
    if set(cols) != expected_cols:
        raise ValueError("The __footings__tab does not contain the correct columns.")
    data = list(data)
    df_log = pd.DataFrame(data, columns=cols)

    values = {}
    for record in df_log.itertuples():
        ws = wb[record.worksheet]
        if "DataFrame" in str(record.dtype) or "Series" in str(record.dtype):
            records = []
            col_range = range(record.col_start, record.col_end + 1)
            col_headers = [ws.cell(record.row_start, c).value for c in col_range]
            for row in range(record.row_start + 1, record.row_end + 1):
                row_capture = {
                    header: ws.cell(row, col).value
                    for header, col in zip(col_headers, col_range)
                }
                records.append(row_capture)
            values.update({_make_key(record): pd.DataFrame.from_records(records)})
        else:
            values.update(
                {_make_key(record): ws.cell(record.row_start, record.col_start).value}
            )

    return values



from datetime import date
from typing import Union

import pandas as pd

from footings.utils import dispatch_function


@dispatch_function(key_parameters=("type_birth_dt", "type_as_of_dt",))
def _calculate_age(type_birth_dt, type_as_of_dt, birth_dt, as_of_dt, method):
    msg = "No registered function based on passed paramters and no default function."
    raise NotImplementedError(msg)


def calculate_age(
    birth_dt: Union[date, pd.Timestamp, pd.Series],
    as_of_dt: Union[date, pd.Timestamp, pd.Series],
    method: str,
) -> Union[int, pd.Series]:
    """
    Calculate age given a birth date and an as of date.

    The calculate age accepts many different types for both the birth_dt and as_of_dt. Thus, this
    formula can be used to calculate the age difference between a single date as well as a pandas
    series of dates.

    Parameters
    ----------
    birth_dt : date or pd.Timestamp or pd.Series
        The birth date.
    as_of_dt : date or pd.Timestamp or pd.Series
        The as of date to calculate age.
    method : str
        Options are - \n
        - ALB = Age last birthday \n
        - ACB = Age closest birthday \n
        - ANB = Age next birthday \n

    Returns
    -------
    int or pd.Series
        Returns int if both birth_dt and as_of_dt are not pd.Series. Returns a pd.Series if any
        parameter is a pd.Series.

    Examples
    --------
    >>> birth_dt = date(1990, 5, 30)
    >>> as_of_dt = date(2020, 4, 6)
    >>> ts_birth_dt = pd.Timestamp("1990-05-30")
    >>> ts_as_of_dt = pd.Timestamp("2020-04-06")
    >>> pd_birth_dt = pd.Series([date(1990, 5, 30)], dtype="datetime64[ns]")
    >>> pd_as_of_dt = pd.Series([date(2020, 4, 6)], dtype="datetime64[ns]")
    >>>
    >>> calculate_age(pd_birth_dt, pd_as_of_dt, "ALB") # pd.Series([29])
    >>> calculate_age(pd_birth_dt, as_of_dt, "ANB") # pd.Series([30])
    >>> calculate_age(pd_birth_dt, ts_as_of_dt, "ACB") # pd.Series([30])
    >>>
    >>> calculate_age(birth_dt, as_of_dt, "ALB") # 29
    >>> calculate_age(birth_dt, pd_as_of_dt, "ANB") # pd.Series([30])
    >>> calculate_age(birth_dt, ts_as_of_dt, "ACB") # 30
    >>>
    >>> calculate_age(ts_birth_dt, ts_as_of_dt, "ALB") # 29
    >>> calculate_age(ts_birth_dt, pd_as_of_dt, "ANB") # pd.Series([30])
    >>> calculate_age(ts_birth_dt, as_of_dt, "ACB") # 30
    """
    return _calculate_age(
        type_birth_dt=type(birth_dt),
        type_as_of_dt=type(as_of_dt),
        birth_dt=birth_dt,
        as_of_dt=as_of_dt,
        method=method,
    )


def _series_calculate_age(birth_dt, as_of_dt, method):
    diff = (as_of_dt.year - birth_dt.year) * 12 + (as_of_dt.month - birth_dt.month)
    if method == "ALB":
        return diff.floordiv(12).astype("int32")
    if method == "ANB":
        return diff.floordiv(12).add(1).astype("int32")
    if method == "ACB":
        return diff.div(12).round(0).astype("int32")
    raise ValueError(f"The method passed [{method}] is not known. See the documentation.")


def _base_calculate_age(birth_dt, as_of_dt, method):
    diff = (as_of_dt.year - birth_dt.year) * 12 + (as_of_dt.month - birth_dt.month)
    if method == "ALB":
        return diff // 12
    if method == "ANB":
        return diff // 12 + 1
    if method == "ACB":
        return int(round(diff / 12, 0))
    raise ValueError(f"The method passed [{method}] is not known. See the documentation.")


@_calculate_age.register(type_birth_dt=pd.Series, type_as_of_dt=pd.Series)
def _(birth_dt, as_of_dt, method):
    return _series_calculate_age(birth_dt.dt, as_of_dt.dt, method)


@_calculate_age.register(type_birth_dt=pd.Series, type_as_of_dt=date)
def _(birth_dt, as_of_dt, method):
    return _series_calculate_age(birth_dt.dt, as_of_dt, method)


@_calculate_age.register(type_birth_dt=pd.Series, type_as_of_dt=pd.Timestamp)
def _(birth_dt, as_of_dt, method):
    return _series_calculate_age(birth_dt.dt, as_of_dt, method)


@_calculate_age.register(type_birth_dt=date, type_as_of_dt=date)
def _(birth_dt, as_of_dt, method):
    return _base_calculate_age(birth_dt, as_of_dt, method)


@_calculate_age.register(type_birth_dt=date, type_as_of_dt=pd.Series)
def _(birth_dt, as_of_dt, method):
    return _series_calculate_age(birth_dt, as_of_dt.dt, method)


@_calculate_age.register(type_birth_dt=date, type_as_of_dt=pd.Timestamp)
def _(birth_dt, as_of_dt, method):
    return _base_calculate_age(birth_dt, as_of_dt, method)


@_calculate_age.register(type_birth_dt=pd.Timestamp, type_as_of_dt=pd.Timestamp)
def _(birth_dt, as_of_dt, method):
    return _base_calculate_age(birth_dt, as_of_dt, method)


@_calculate_age.register(type_birth_dt=pd.Timestamp, type_as_of_dt=pd.Series)
def _(birth_dt, as_of_dt, method):
    return _series_calculate_age(birth_dt, as_of_dt.dt, method)


@_calculate_age.register(type_birth_dt=pd.Timestamp, type_as_of_dt=date)
def _(birth_dt, as_of_dt, method):
    return _base_calculate_age(birth_dt, as_of_dt, method)



import sys
from traceback import extract_tb, format_list

from attr import asdict, attrib, attrs
from attr.validators import instance_of


@attrs(frozen=True, slots=True)
class Error:
    """An exception as an object.

    :param key: The key identifying the point of error.
    :param error_type: The error type.
    :param error_value: The error value.
    :param error_stacktrace: The stacktrace of the error.
    """

    key = attrib(type=str, validator=instance_of(str))
    error_type = attrib(type=str, validator=instance_of(str))
    error_value = attrib(type=str, validator=instance_of(str))
    error_stacktrace = attrib(type=str, validator=instance_of(str))

    @classmethod
    def create(cls, key, sys_info):
        exc_type, exc_value, exc_trace = sys.exc_info()
        return cls(
            key=str(key),
            error_type=exc_type.__qualname__,
            error_value=str(exc_value.args),
            error_stacktrace=str(format_list(extract_tb(exc_trace))),
        )

    def to_audit_json(self):
        return asdict(self)

    def to_audit_xlsx(self):
        return asdict(self)


class ModelRunError(Exception):
    """Error occured while running a footings model."""


class ModelCreationError(Exception):
    """Error occured when creating a footings model."""


class DataDictionaryValidatorsConversionError(Exception):
    """Error occured when converting validator to Column."""


class DataDictionaryPandasDtypeConversionError(Exception):
    """Error occured when converting object to PandasDtype."""


class DataDictionaryValidateError(Exception):
    """Validator failed when testing whether dataframe is valid."""



from .age import calculate_age
from .calculations import (  # calc_change_in_reserve,
    calc_benefit_reserve,
    calc_continuance,
    calc_discount,
    calc_interpolation,
    calc_pv,
    calc_pvfnb,
)
from .frames import (
    create_frame,
    create_frame_from_record,
    expand_frame_per_record,
    frame_add_exposure,
    frame_add_weights,
    frame_filter,
)
from .meta_tools import run_date_time
from .records import convert_to_records



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['footings',
 'footings.actuarial_tools',
 'footings.doc_tools',
 'footings.io',
 'footings.parallel_tools']

package_data = \
{'': ['*'], 'footings.doc_tools': ['templates/*']}

install_requires = \
['attrs>=20.0,<21.0',
 'numpy>=1.11,<2.0',
 'numpydoc>=1.0,<2.0',
 'openpyxl>=3.0,<4.0',
 'pandas>=1.0,<2.0']

setup_kwargs = {
    'name': 'footings',
    'version': '0.10.0',
    'description': 'A model building library',
    'long_description': '# Footings - A Model Building Library\n\n![tests](https://github.com/footings/footings/workflows/tests/badge.svg)\n[![gh-pages](https://github.com/footings/footings/workflows/gh-pages/badge.svg)](https://footings.github.io/footings/master/)\n[![license](https://img.shields.io/badge/License-BSD%203--Clause-blue.svg)](https://opensource.org/licenses/BSD-3-Clause)\n[![codecov](https://codecov.io/gh/dustindall/footings/branch/master/graph/badge.svg?token=SC5BHMYBSN)](https://codecov.io/gh/dustindall/footings)\n![version](https://img.shields.io/pypi/pyversions/footings.svg)\n[![PyPI version](https://badge.fury.io/py/footings.svg)](https://badge.fury.io/py/footings)\n\n## What is it?\n\nFootings is a model building Python library. No out-of-the box models are provided. Instead it is a framework library that provides key objects and functions to help users  construct custom models.\n\n## Purpose\n\nThe footings library was developed with the intention of making it easier to develop actuarial models in Python. Actuarial models are a mix of data engineering and math/calculations. In addition, actuarial models are usually not defined by one calculation, but a series of calculations. So even though the original purpose is actuarial work, if the problem at hand sounds familiar, others might find this library useful.\n\n## Principles\n\nThe Footings library was designed as framework library using the below principles -\n\n- Models are a sequence of linked steps\n- Models need to be easy to understand\n- Models need to have validation built in\n- Models need to be easy to audit\n- Models need to be self documenting\n- Models need to be able to scale up\n- Models need to be able to build off other models\n- Model environments should not be monolithic\n\n**These all become easier when you can leverage the amazing Python and wider open source ecosystems.**\n\n## Installation\n\nTo install from PyPI run -\n\n```\npip install footings\n```\n\nTo install the latest version on github run -\n\n```\npip install git+https://github.com/foootings/footings-core.git\n```\n\n## License\n[BSD 3](LICENSE)\n\n\n## Changelog\n\n[File](./docs/changelog.md)\n',
    'author': 'Dustin Tindall',
    'author_email': 'dustin.tindall@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/footings/footings',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import json
from operator import add
from typing import Mapping


def flatten_dict(d):
    results = []

    def lift(x):
        return f"/{str(x)}"

    def visit(subdict, results, partial_key=None):
        for k, v in subdict.items():
            new_key = f"/{str(k)}" if partial_key is None else add(partial_key, lift(k))
            if isinstance(v, Mapping):
                visit(v, results, new_key)
            else:
                results.append((new_key, v))

    visit(d, results, None)
    return dict(results)


def load_footings_json_file(file: str):
    """Load footings generated json file.

    Parameters
    ----------
    file : str
        The path to the file.

    Returns
    -------
    dict
        A dict representing the respective file type.
    """
    with open(file, "r") as f:
        loaded_json = json.load(f)
    return flatten_dict(loaded_json)
